/*
 * Copyright (c) 2006 - 2007, Nils R. Weller
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * x86 backend
 * (XXX much of this stuff can probably be adapted to different
 * architectures)
 */
#include "amd64_gen.h"
#include "backend.h"
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>
#include <limits.h>
#include "scope.h"
#include "decl.h"
#include "type.h"
#include "decl.h"
#include "icode.h"
#include "functions.h"
#include "control.h"
#include "debug.h"
#include "token.h"
#include "error.h"
#include "functions.h"
#include "symlist.h"
#include "icode.h"
#include "stack.h"
#include "reg.h"
#include "subexpr.h"
#include "expr.h"
/* #include "x86_emit_gas.h" */
#include "inlineasm.h"
#include "x86_emit_nasm.h"
#include "x86_emit_gas.h"
#include "x86_gen.h"
#include "amd64_emit_yasm.h"
#include "amd64_emit_gas.h"
#include "cc1_main.h"
#include "n_libc.h"

static FILE			*out;
static struct scope		*tunit;
static int			use_nasm = 1; /* XXX */

static int			rbx_saved;
struct vreg			csave_rbx;
struct emitter_amd64		*emit_amd64;


#define N_GPRS	6	
#define N_ARGREGS 6 

struct reg		*amd64_argregs[] = {
	/* rdi, rsi, rdx, rcx, r8, r9 */
	&amd64_x86_gprs[5], &amd64_x86_gprs[4],
	&amd64_x86_gprs[3], &amd64_x86_gprs[2],
	&amd64_gprs[8], &amd64_gprs[9]
};

#if 0
static struct reg	amd64_rsp;
static struct reg	amd64_rbp;
#endif

struct reg		amd64_x86_gprs[7];
struct reg		amd64_gprs[16];
struct reg		amd64_gprs_32bit[16];
struct reg		amd64_gprs_16bit[16];
struct reg		amd64_gprs_8bit[16];
struct reg		amd64_sil;
struct reg		amd64_dil;
struct reg		amd64_sse_regs[8];

#if 0
struct reg		amd64_spl;
struct reg		amd64_bpl;
#endif

static int		callee_save_map[] = {
	0, 0, 0, 0, /* r8 - r11 */
	1, 1, 1, 1 /* r12 - r15 */
};		


static void
init_regs(void) {
	static struct reg	nullreg;
	int			i, j;
	static const struct {
		struct reg	*regs;
		char		*names[9];
	} rps[] = {
		{ amd64_x86_gprs,
			{"rax","rbx","rcx","rdx","rsi","rdi",0,0,0}},
		{ NULL, {0,0,0,0,0,0,0,0,0} }
	};
	
	for (i = 0; rps[i].regs != NULL; ++i) {
		nullreg.type = REG_GPR;
		nullreg.allocatable = 1;
		for (j = 0; rps[i].names[j] != NULL; ++j) {
			rps[i].regs[j] = nullreg;
			rps[i].regs[j].composed_of =
				n_xmalloc(2 * sizeof(struct reg *));
			rps[i].regs[j].composed_of[0] = &x86_gprs[j];
			rps[i].regs[j].composed_of[1] = NULL;
			rps[i].regs[j].size = 8;
			rps[i].regs[j].name = rps[i].names[j];
		}
	}
	
	amd64_sil.size = 1;
	amd64_sil.name = "sil";
	amd64_sil.type = REG_GPR;
	amd64_sil.allocatable = 1;
	x86_gprs[4].composed_of[0]->composed_of =
		n_xmalloc(2 * sizeof(struct reg *));
	x86_gprs[4].composed_of[0]->composed_of[0] = &amd64_sil;
	x86_gprs[4].composed_of[0]->composed_of[1] = NULL;
	
	amd64_dil.size = 1;
	amd64_dil.name = "dil";
	amd64_dil.type = REG_GPR;
	amd64_dil.allocatable = 1;
	x86_gprs[5].composed_of[0]->composed_of =
		n_xmalloc(2 * sizeof(struct reg *));
	x86_gprs[5].composed_of[0]->composed_of[0] = &amd64_dil;
	x86_gprs[5].composed_of[0]->composed_of[1] = NULL;
	
	for (i = 8; i < 16; ++i) {
		static char	*new_gpr_names[] = {
			"r8", "r9", "r10", "r11",
			"r12", "r13", "r14", "r15"
		};
		static char	*new_gpr_names_32[] = {
			"r8d", "r9d", "r10d", "r11d",
			"r12d", "r13d", "r14d", "r15d"
		};
		static char	*new_gpr_names_16[] = {
			"r8w", "r9w", "r10w", "r11w",
			"r12w", "r13w", "r14w", "r15w"
		};
		static char	*new_gpr_names_8[] = {
			"r8b", "r9b", "r10b", "r11b",
			"r12b", "r13b", "r14b", "r15b"
		};
		amd64_gprs[i].name = new_gpr_names[i-8];
		amd64_gprs[i].size = 8;
		amd64_gprs[i].type = REG_GPR;
		amd64_gprs[i].allocatable = 1;
		amd64_gprs[i].composed_of = n_xmalloc(2 * sizeof(struct reg*));
		amd64_gprs[i].composed_of[0] = &amd64_gprs_32bit[i];
		amd64_gprs[i].composed_of[1] = NULL;

		amd64_gprs_32bit[i].name = new_gpr_names_32[i-8];
		amd64_gprs_32bit[i].size = 4;
		amd64_gprs_32bit[i].type = REG_GPR;
		amd64_gprs_32bit[i].allocatable = 1;
		amd64_gprs_32bit[i].composed_of
			= n_xmalloc(2 * sizeof(struct reg*));
		amd64_gprs_32bit[i].composed_of[0] = &amd64_gprs_16bit[i];
		amd64_gprs_32bit[i].composed_of[1] = NULL;
		
		amd64_gprs_16bit[i].name = new_gpr_names_16[i-8];
		amd64_gprs_16bit[i].size = 2;
		amd64_gprs_16bit[i].type = REG_GPR;
		amd64_gprs_16bit[i].allocatable = 1;
		amd64_gprs_16bit[i].composed_of
			= n_xmalloc(2 * sizeof(struct reg*));
		amd64_gprs_16bit[i].composed_of[0] = &amd64_gprs_8bit[i];
		amd64_gprs_16bit[i].composed_of[1] = NULL;

		amd64_gprs_8bit[i].name = new_gpr_names_8[i-8];
		amd64_gprs_8bit[i].size = 1;
		amd64_gprs_8bit[i].type = REG_GPR;
		amd64_gprs_8bit[i].allocatable = 1;
		amd64_gprs_8bit[i].composed_of = NULL;
	}
	
	amd64_x86_gprs[6].name = NULL;

	for (i = 0; i < 8; ++i) {
		static char	*names[] = {
			"xmm0", "xmm1", "xmm2", "xmm3",
			"xmm4", "xmm5", "xmm6", "xmm7",
		};
		amd64_sse_regs[i].name = names[i];
		amd64_sse_regs[i].type = REG_FPR;
		amd64_sse_regs[i].size = 8; /* XXX */
		amd64_sse_regs[i].allocatable = 1;
	}
}

struct reg *
find_top_reg(struct reg *r) {
	int	i;

	for (i = 0; i < 6; ++i) {
		if (is_member_of_reg(&amd64_x86_gprs[i], r)) {
			return &amd64_x86_gprs[i];
		}
	}
	abort();
	return NULL;
}


#if 0

#define MEM_CLASS_NONE		1
#define MEM_CLASS_MEMORY	2	
#define MEM_CLASS_INTEGER	3	
#define MEM_CLASS_X87		4
#define MEM_CLASS_X87UP		5
#define MEM_CLASS_COMPLEX_X87	6
#define MEM_CLASS_SSE		7

static int
classify_memory_class(struct type *ty, int *second_class) {
	if (is_integral_type(ty)
		|| ty->tlist != NULL) {
		return MEM_CLASS_INTEGER;
	} else if (IS_FLOATING(ty->code)) {
		/* XXX */
		return MEM_CLASS_SSE;
	} else if (ty->code == TY_STRUCT
		|| ty->code == TY_UNION) {
		struct sym_entry	*se;
		int			rc = MEM_CLASS_NONE;
		int			cur_class = MEM_CLASS_NONE;
		int			field_class;

		if (ty->tstruc->size > 16) {
			return MEM_CLASS_MEMORY;
		}

		se = ty->tstruc->scope->slist;
		for (; se != NULL; se = se->next) {
			if (se->offset >= 8) {
				rc = cur_class;
				cur_class = MEM_CLASS_NONE;
			}
			field_class = classify_memory_class(

		}
		return rc;
	}
} 

#endif

static void
do_invalidate(struct reg *r, struct icode_list *il, int save) {
	free_preg(r, il, 1, save);
}

/*
 * XXX this shouldn't be saving esi/edi/ebx and r12 - r15 when we're
 * invalidating because of a function call, because those regs are
 * callee-save
 */
static void
invalidate_gprs(struct icode_list *il, int saveregs) {
	int	i;

	for (i = 0; i < N_GPRS; ++i) {
		do_invalidate(&amd64_x86_gprs[i], il, saveregs);
	}
	for (i = 8; i < 16; ++i) {
		do_invalidate(&amd64_gprs[i], il, saveregs);
	}	
}


static struct reg *
alloc_gpr(struct function *f, int size, struct icode_list *il,
struct reg *dontwipe, int line) {
	struct reg	*ret;

	if (size == 0) {
		/* 0 means GPR */
		size = 8;
	}

	if (size < 8) {
		ret = x86_backend.alloc_gpr(f, size, il, dontwipe, line);
	} else {
		/*
		 * Notice how only r8 - r15 are used for 64bit register
		 * allocations. This is because the x86 gpr extensions
		 * (rax, rbx, etc) are used for argument passing, so
		 * thrashing should be avoided. Note that emit_copystruct()
		 * will use those regs too, so it is absolutely critical	
		 * that struct pointers used by it are never stored in
		 * rdi/rsi/rdx
		 */
		ret = generic_alloc_gpr(f, size, il, dontwipe,
			&amd64_gprs[8], 8, callee_save_map, line);
	}

	return ret;
}

static int	sse_csave_map[] = {
	0, 0, 0, 0, 0, 0, 0, 0
};

static struct reg *
alloc_fpr(struct function *f, int size, struct icode_list *il,
struct reg *dontwipe) {

#if 0
	if ((unsigned)size == /*12*/
		backend->get_sizeof_type(make_basic_type(TY_LDOUBLE), 0)) {
#endif
	/*
	 * What the HELL IS GOING ON HERE!!!!!!!!!!!!!!
	 * ANOTHER 10 vs 12 CONFLICT, I AM BECOMING INSANE!!!!!!!!
	 */
	if (size == 10 || size == 12 || size == 16) {
		return x86_backend.alloc_fpr(f, size, il, dontwipe);
	} else {
		return generic_alloc_gpr(f, size, il, dontwipe,
			amd64_sse_regs, 8, sse_csave_map, 0); 
	}
}

static void
x86_free_preg(struct reg *r, struct icode_list *il) {
	x86_backend.free_preg(r, il);
}

/*
 * IMPORTANT: The x86 backend and the x86 emitter that corresponds to
 * this emitter (currently only yasm) also has to be initialized (init()),
 * because some code is shared between x86 and amd64
 */
static int 
init(FILE *fd, struct scope *s) {
	out = fd;
	tunit  = s;

	(void) use_nasm;
	if (asmflag == NULL
		|| strcmp(asmflag, "yasm") == 0) {
		/* Default is yasm */
		emit = &amd64_emit_yasm;
		emit_x86 = &x86_emit_x86_nasm; /* XXX */
		x86_backend.init(fd, s);
		x86_emit_nasm.init(out, tunit);
		emit_amd64 = &emit_amd64_yasm;
	} else if (strcmp(asmflag, "gas") == 0) {
		emit = &amd64_emit_gas;
		emit_x86 = &x86_emit_x86_gas;
		x86_backend.init(fd, s);
		x86_emit_gas.init(out, tunit);
		emit_amd64 = &emit_amd64_gas;
	} else {
		(void) fprintf(stderr, "Unknown AMD64 assembler `%s'\n",
			asmflag);
		exit(EXIT_FAILURE);
	}

	init_regs();

	/* Setup code sharing between x86 and amd64 */
	amd64_backend.invalidate_except = x86_backend.invalidate_except;
	amd64_backend.name_to_reg = x86_backend.name_to_reg;
	amd64_backend.get_inlineasm_label = x86_backend.get_inlineasm_label;
	amd64_backend.asmvreg_to_reg = x86_backend.asmvreg_to_reg;
	amd64_backend.alloc_16_or_32bit_noesiedi =
		x86_backend.alloc_16_or_32bit_noesiedi;
	backend->emit = emit;
	return emit->init(out, tunit);
}

static int
get_ptr_size(void) {
	return 8;
}	

static struct type *
get_size_t(void) {
	return make_basic_type(TY_ULONG);
}	

static size_t
get_sizeof_basic(int type) {
	switch (type) {
	case TY_ENUM:
		return 4; /* XXX */

	case TY_INT:
	case TY_UINT:
		return 4;
	case TY_LONG:
	case TY_ULONG:
	case TY_LLONG:
	case TY_ULLONG:
		return 8;

	case TY_CHAR:
	case TY_UCHAR:
	case TY_SCHAR:
	case TY_BOOL:
		return 1;

	case TY_SHORT:
	case TY_USHORT:
		return 2;

	case TY_FLOAT:
		return 4;

	case TY_DOUBLE:
		return 8; /* XXX contradicts abi */

	case TY_LDOUBLE:
		return /*10*/12;
	default:
	printf("err sizeof cannot cope w/ it, wuz %d\n", type); 
	abort();
		return 1; /* XXX */
	}
}

static struct vreg		saved_gprs[4]; /* r12 - r15 */
static struct stack_block	*saved_gprs_sb[4];

static void
do_ret(struct function *f, struct icode_instr *ip) {
	int	i;

	if (f->alloca_head != NULL) {
		struct stack_block	*sb;
		static struct vreg	rvr;

		rvr.stack_addr = f->alloca_regs;
		rvr.size = 8;
		vreg_map_preg(&rvr, &amd64_x86_gprs[0]);
		emit->store(&rvr, &rvr);

		for (sb = f->alloca_head; sb != NULL; sb = sb->next) {
			emit->dealloca(sb);
		}

		emit->load(&amd64_x86_gprs[0], &rvr);
	}
	if (f->callee_save_used & CSAVE_EBX) {
		emit->load(&amd64_x86_gprs[1], &csave_rbx);
	}
	for (i = 12; i < 16; ++i) {
		if (saved_gprs[i-12].stack_addr != NULL) {
			emit->load(&amd64_gprs[i], &saved_gprs[i-12]);
		}
	}

	if (saved_ret_addr) {
		emit->check_ret_addr(f, saved_ret_addr);
	}
	emit->freestack(f, NULL);
	emit->ret(ip);
}

static void
map_parameters(struct function *f, struct ty_func *proto) {
	struct sym_entry	*se = proto->scope->slist;
	struct stack_block	*sb;
	int			i;
	long			offset = 16; /* rbp */
	int			gprs_used = 0;
	int			fprs_used = 0;
	struct reg		*curreg;

	if (f->fty->variadic) {
		/*
		 * Same story as usual - allocate space for argument
		 * registers; those are then followed by any possibly
		 * stack-passed variadic arguments.
		 * 6 arg regs * 8 = 48 bytes
		 * XXX floating point!! mmm..sse registers.mmm
		 */
		f->fty->lastarg = alloc_decl();

		/* Allocate 48 bytes for gprs, followed by 64 for fprs */
		f->fty->lastarg->stack_addr = stack_malloc(f, /*48*/112);
	}

	if (f->proto->dtype->tlist->next == NULL
		&& (f->proto->dtype->code == TY_STRUCT
		|| f->proto->dtype->code == TY_UNION)) {
		/*
		 * Function returns struct/union - accomodate for
		 * hidden pointer (passed as first argument)
		 * XXX duplicates mips code
		 */
		struct vreg	*hp;
		hp = vreg_alloc(NULL,NULL,NULL,NULL);
		hp->size = 8;
		hp->var_backed = alloc_decl();
		hp->var_backed->dtype =
			n_xmemdup(f->proto->dtype, sizeof(struct type));
		hp->var_backed->dtype->tlist = alloc_type_node();
		hp->var_backed->dtype->tlist->type = TN_POINTER_TO; 
		hp->var_backed->stack_addr = stack_malloc(f, 8);
		f->hidden_pointer = hp;
		++gprs_used;
	}	

	for (i = 0; i < proto->nargs; ++i, se = se->next) {
		size_t		size;

		size = backend->get_sizeof_type(se->dec->dtype, NULL);
#if 0
		sb = make_stack_block(offset, size); 
		offset += size; /* was before makestackblock */
		sb->is_func_arg = 1;
#endif
		if (is_integral_type(se->dec->dtype)
			|| se->dec->dtype->tlist) {
			if (gprs_used < N_ARGREGS) {
				/* passed in register */
				curreg = amd64_argregs[gprs_used++];
				sb = stack_malloc(f, size);
				se->dec->stack_addr = sb;
				if (size == 4) {
					curreg = curreg->composed_of[0];
				} else if (size == 2) {
					curreg = curreg->composed_of[0]
						->composed_of[0];
				} else if (size == 1) {
					if (curreg->composed_of[0]
						->composed_of[0]
						->composed_of[1]) {
						curreg = curreg->composed_of[0]
							->composed_of[0]
							->composed_of[1];
					} else {
						curreg = curreg->composed_of[0]
							->composed_of[0]
							->composed_of[0];
					}
				}
				
				se->dec->stack_addr->from_reg = curreg;
			} else {
				/* passed on stack */
				/* XXX alignment */
				se->dec->stack_addr =
					make_stack_block(offset,
						se->dec->vreg->size);
				se->dec->stack_addr->is_func_arg = 1;
				offset += se->dec->vreg->size;
				while (offset % 8) {
					++offset;
				}
			}
		} else if (IS_FLOATING(se->dec->dtype->code)) {
			if (se->dec->dtype->code == TY_LDOUBLE) {
				sb = make_stack_block(offset, size);
				sb->is_func_arg = 1;
				offset += se->dec->vreg->size;
				while (offset % 8) ++offset;
				se->dec->stack_addr = sb;
			} else {
				if (fprs_used < 8) {
					/* passed in register */
					curreg = &amd64_sse_regs[fprs_used++];
					sb = stack_malloc(f, size);
					se->dec->stack_addr = sb;
					sb->from_reg = curreg;
				} else {
					/* passed on stack */
					/* XXX alignment */
					se->dec->stack_addr =
						make_stack_block(offset,
							se->dec->vreg->size);
					se->dec->stack_addr->is_func_arg = 1;
					offset += se->dec->vreg->size;
					while (offset % 8) {
						++offset;
					}
				}
			}
		} else if (se->dec->dtype->code == TY_STRUCT
			|| se->dec->dtype->code == TY_UNION) {
			if (1 /*size > 16  || has_unaligned_members() */) {
				sb = make_stack_block(offset, size); 
				offset += size; /* was before makestackblock */
				while (offset % 8) ++offset;
				sb->is_func_arg = 1;
				se->dec->stack_addr = sb;
			}
		} else {
			unimpl();
		}
	}
	if (f->fty->variadic) {
		/* Patch varargs block to real address */
		struct stack_block	*save_area;

		save_area = f->fty->lastarg->stack_addr;
		if (gprs_used == 6) {
			/* All variadic stuff passed on stack */
			f->fty->lastarg->stack_addr =
				make_stack_block(offset, 0);
			f->fty->lastarg->stack_addr->is_func_arg = 1;
		} else {
#if 0 
			f->fty->lastarg->stack_addr->offset -=
				gprs_used * 8;
#endif
			f->fty->lastarg->stack_addr->from_reg =
				(void *)&amd64_argregs[gprs_used]; /* XXX */
		}
		if (f->patchme) {
			struct amd64_va_patches	*p = f->patchme;
			int			n;

			if (gprs_used == 6) {
				n = 48;
			} else {
				n = (&amd64_argregs[gprs_used] -
					amd64_argregs) * 8;
			}
			*p->gp_offset = n;
			if (fprs_used == 8) {
				n = 64+48;
			} else {
				n = (&amd64_sse_regs[fprs_used] -
					amd64_sse_regs) * 8;
				n += 48;
			}
			*p->reg_save_area = *save_area;
			if (gprs_used == 6) {
				*p->overflow_arg_area =
					*f->fty->lastarg->stack_addr;
			} else {
				*p->overflow_arg_area =
					*make_stack_block(16, 0);
				p->overflow_arg_area->is_func_arg = 1;
			}
#if 0
printf("gp offset = %d\n", n);
printf("reg save area = %d\n", p->reg_save_area->offset);
printf("overflow area = %d\n", p->overflow_arg_area->offset);
#endif
		}
	}
}

void	store_preg_to_var(struct decl *, size_t, struct reg *);

static int
gen_function(struct function *f) {
	struct ty_func		*proto;
	struct scope		*scope;
	struct icode_instr	*lastret = NULL;
	struct stack_block	*sb;
	struct sym_entry	*se;
	size_t			size;
	size_t			alloca_bytes = 0;
	int			i;
	unsigned		mask;

	proto = f->proto->dtype->tlist->tfunc;

	emit->label(f->proto->dtype->name, 1);
	emit->intro(f);

	map_parameters(f, proto);

	/* Make local variables */
	for (scope = f->scope; scope != NULL; scope = scope->next) {
		struct stack_block	*sb;
		struct scope		*tmp;
		struct decl		**dec;
		size_t			align;

		for (tmp = scope; tmp != NULL; tmp = tmp->parent) {
			if (tmp == f->scope) {
				break;
			}
		}

		if (tmp == NULL) {
			/* End of function reached */
			break;
		}
		if (scope->type != SCOPE_CODE) continue;

		dec = scope->automatic_decls.data;
		for (i = 0; i < scope->automatic_decls.ndecls; ++i) {
			if (dec[i]->stack_addr != NULL) { /* XXX sucks */
				continue;
			}

			if (i+1 < scope->automatic_decls.ndecls) {
				align = calc_align_bytes(f->total_allocated,
					dec[i]->dtype,
					dec[i+1]->dtype);	
			} else {
				align = 0;
			}	

			size = backend->
				get_sizeof_decl(dec[i], NULL);
			sb = stack_malloc(f, size+align);
			sb->nbytes = size;
			dec[i]->stack_addr = sb;
		}
	}
	stack_align(f, 8);

	/*
	 * Allocate storage for saving callee-saved registers (ebx/esi/edi)
	 * (but defer saving them until esp has been updated)
	 */
	f->total_allocated += 8;
	if (f->callee_save_used & CSAVE_EBX) {
		rbx_saved = 1;
		csave_rbx.stack_addr
			= make_stack_block(f->total_allocated, 8);
	}	

	for (i = 12, mask = 1 << 11; i < 16; ++i, mask <<= 1) {
		if (f->callee_save_used & mask) {
			if (saved_gprs_sb[i-12] == NULL) {
				saved_gprs_sb[i-12] = make_stack_block(0, 8);
			}	
			f->total_allocated += 8;
			saved_gprs[i-12].stack_addr = saved_gprs_sb[i-12];
			saved_gprs[i-12].size = 8;
			saved_gprs[i-12].stack_addr->offset =
				f->total_allocated;
		} else {
			saved_gprs[i-12].stack_addr = NULL;
		}	
	}
	f->callee_save_offset = f->total_allocated;

	if (stackprotectflag) {
		f->total_allocated += 4;
		if (saved_ret_addr == NULL) {
			saved_ret_addr
				= make_stack_block(f->total_allocated, 4);
		} else {
			saved_ret_addr->offset = f->total_allocated;
			saved_ret_addr->nbytes = 4;
		}	
	}

	/* Allocate storage for temporarily saving GPRs & patch offsets */
	for (sb = f->regs_head; sb != NULL; sb = sb->next) {
		stack_align(f, sb->nbytes);
		f->total_allocated += sb->nbytes;
		sb->offset = f->total_allocated;
	}
	/*
	 * Allocate storage for saving alloca() pointers, and initialize
	 * it to zero
	 */
	stack_align(f, 8);
	for (sb = f->alloca_head; sb != NULL; sb = sb->next) {
		f->total_allocated += sb->nbytes;
		alloca_bytes += sb->nbytes;
		sb->offset = f->total_allocated;
	}
	if (f->alloca_head != NULL) {
		/*
		 * Get stack for saving return value register (rax)
		 * before performing free() on alloca()ted blocks
		 */
		f->alloca_regs = make_stack_block(0, 8);
		f->total_allocated += 8;
		f->alloca_regs->offset = f->total_allocated;
	}

	if (f->total_allocated > 0) {
		stack_align(f, 16);
		emit->allocstack(f, f->total_allocated);
		if (f->callee_save_used & CSAVE_EBX) {
			vreg_map_preg(&csave_rbx, &amd64_x86_gprs[1]);
			emit->store(&csave_rbx, &csave_rbx);
			x86_gprs[1].used = 0;
			amd64_gprs[1].used = 0;
		}
		for (i = 0; i < 4; ++i) {
			if (saved_gprs[i].stack_addr != NULL) {
				vreg_map_preg(&saved_gprs[i],
					&amd64_gprs[12+i]);
				emit->store(&saved_gprs[i], &saved_gprs[i]);
				amd64_gprs[12+i].used = 0;
			}
		}	
		if (f->hidden_pointer) {
			vreg_map_preg(f->hidden_pointer, &amd64_x86_gprs[5]);
			emit->store(f->hidden_pointer, f->hidden_pointer);
		}
		se = proto->scope->slist;
		for (i = 0; i < proto->nargs; ++i, se = se->next) {
			if (se->dec->stack_addr->from_reg != NULL) {
				vreg_map_preg(se->dec->vreg,
					se->dec->stack_addr->from_reg);
				emit->store(se->dec->vreg, se->dec->vreg);
			}
		}
		if (f->fty->variadic
			&& f->fty->lastarg->stack_addr->from_reg != NULL) {
			struct reg	**r;
			size_t		saved_offset =
				f->fty->lastarg->stack_addr->offset;

			r = (struct reg **)f->fty->
				lastarg->stack_addr->from_reg; /* XXX */
			f->fty->lastarg->stack_addr->offset -=
				(r - amd64_argregs) * 8;
			for (i = r - amd64_argregs; i < N_ARGREGS; ++i) {
				store_preg_to_var(f->fty->lastarg, 8,
					amd64_argregs[i]);
				f->fty->lastarg->stack_addr->offset -= 8;
			}

			/* XXX ... */
			for (i = 0; i < 8; ++i) {
				f->fty->lastarg->dtype =
					make_basic_type(TY_DOUBLE);
				store_preg_to_var(f->fty->lastarg, 8,
					&amd64_sse_regs[i]);
				f->fty->lastarg->stack_addr->offset -= 8;
			}

			f->fty->lastarg->stack_addr->offset = saved_offset;
		}
	}
	if (stackprotectflag) {
		emit->save_ret_addr(f, saved_ret_addr);
	}
	if (curfunc->alloca_head != NULL) {
		emit->zerostack(curfunc->alloca_head, alloca_bytes);
	}	

	if (xlate_icode(f, f->icode, &lastret) != 0) {
		return -1;
	}
	if (lastret != NULL) {
		struct icode_instr	*tmp;

		for (tmp = lastret->next; tmp != NULL; tmp = tmp->next) {
			if (tmp->type != INSTR_SETITEM) {
				lastret = NULL;
				break;
			}
		}
	}

	if (lastret == NULL) { 
		do_ret(f, NULL);
	}
	return 0;
}


static int
gen_program(void) {
	struct function		*func;

	if (gflag) {
		/* Print file names */
		emit->dwarf2_files();
	}

	if (emit->extern_decls) {
		emit->extern_decls();
	}	

	emit->static_decls();
	emit->struct_inits();
	emit->empty();
	emit->strings();
	emit->empty();

	if (emit->struct_defs) {
		emit->struct_defs();
	}	

	emit->setsection(SECTION_TEXT);

	for (func = funclist; func != NULL; func = func->next) {
		curfunc = func;
		if (gen_function(func) != 0) {
			return -1;
		}
		emit->empty();
		emit->empty();
	}
	x_fflush(out);

	return 0;
}

static void
pass_struct_union(
	struct vreg *vr,
	int *gprs_used, int *fprs_used,
	size_t *stack_bytes_used,
	struct icode_list *il) {

	(void) gprs_used; (void) fprs_used;
	if (1 /*vr->size > 16 || has_unaligned_fields(...) */) {
		struct vreg	*dest;

		/*
		 * memcpy() it onto stack, allocate storage
		 * manually (no push!)
		 */
		dest = vreg_stack_alloc(vr->type, il);
		*stack_bytes_used += dest->size;
		while (*stack_bytes_used % 8) ++*stack_bytes_used;
		backend->invalidate_gprs(il, 1);
		vreg_faultin_ptr(vr, il);
		icode_make_copystruct(dest, vr, il);
	} else {
	}
}


static void
pass_ldouble_stack(
	struct vreg *vr,
	unsigned long *allpushed,
	struct icode_list *il) {
	struct vreg	*dest;
	unsigned long	oldpushed = *allpushed;

	while ((*allpushed + vr->size) % 8) ++*allpushed;
	if (*allpushed != oldpushed) {
		icode_make_allocstack(NULL, *allpushed - oldpushed, il);
	}
	dest = vreg_stack_alloc(vr->type, il);
	*allpushed += vr->size;
	vreg_faultin_x87(NULL, NULL, vr, il, 0);
	vreg_map_preg(dest, vr->pregs[0]);
	icode_make_store(curfunc, dest, dest, il);
}

static unsigned long
pass_args_stack(struct vreg **vrs, int i, int nvrs, struct icode_list *il) {
	int		j;
	unsigned long	allpushed = 0;

	for (j = nvrs - 1; j >= i; --j) {
		int		remaining = 0;
		struct vreg	*dest;
		size_t		tysize = backend->
				get_sizeof_type(vrs[j]->type, NULL);

		/* Pass on stack! (minimum chunk = 8 bytes) */
		while ((allpushed+remaining+tysize) % 8) {
			++remaining;
		}
		if (remaining) {
			icode_make_allocstack(NULL,
				remaining, il);
			allpushed += remaining;
		}
		dest = vreg_stack_alloc(vrs[j]->type, il);
		allpushed += dest->size;
		vreg_faultin(NULL, NULL, vrs[j], il, 0);
		vreg_map_preg(dest, vrs[j]->pregs[0]);
		icode_make_store(curfunc, dest, dest, il);
	}
		/* There may be long double arguments left to pass */
	for (--i; i >= 0; --i) {
		if (vrs[i]->type->code == TY_LDOUBLE
			&& vrs[i]->type->tlist == NULL) {
			pass_ldouble_stack(vrs[i], &allpushed, il);
		}
	}
	return allpushed;
}


#if 0
/*
 * To ensure correct (16-byte) stack alignment, it is necessary to
 * calculate the amount of stack space used for argument passing in
 * advance, and to allocate as much unused padding space as
 * necessary to make it all fit
 */
static unsigend long
calc_stack_bytes(struct vreg **vrs, int nvrs) {
	int	i;

	for (i = 0; i < nvrs; ++i) {
		if (vrs[i]->type
}
#endif


static struct vreg *
icode_make_fcall(struct fcall_data *fcall, struct vreg **vrs, int nvrs,
struct icode_list *il)
{
	size_t			allpushed = 0;
	struct vreg		*tmpvr;
	struct vreg		*ret = NULL;
	struct vreg		*vr2;
	struct type		*ty;
	struct icode_instr	*ii;
	struct type_node	*tn;
	struct vreg		*struct_lvalue;
	int			i;
	int			need_dap = 0;
	int			was_struct;
	int			was_float;
	int			was_llong;
	int			struct_return = 0;
	int			regs_used = 0;
	int			fp_regs_used = 0;
	int			passed_stack = 0;
#if 0
	static struct reg	*amd64_argregs[] = {
		/* rdi, rsi, rdx, rcx, r8, r9 */
		&amd64_x86_gprs[5], &amd64_x86_gprs[4],
		&amd64_x86_gprs[3], &amd64_x86_gprs[2],
		&amd64_gprs[8], &amd64_gprs[9], NULL
	};		
#endif

	ty = fcall->calltovr->type;
	tmpvr = fcall->calltovr;

	tn = ty->tlist;
	if (tn->type == TN_POINTER_TO) {
		/* Called thru function pointer */
		tn = tn->next;
	}

	struct_lvalue = fcall->lvalue;

	if ((ty->code == TY_STRUCT
		|| ty->code == TY_UNION)
		&& tn->next == NULL) {
		struct_return = 1;
		if (struct_lvalue == NULL) {
			struct type_node	*tnsav;
			/*
			 * Result of function is not assigned so we need to
			 * allocate storage for the callee to store its
			 * result into
			 */

			tnsav = ty->tlist;
			ty->tlist = NULL;
			struct_lvalue = vreg_stack_alloc(ty, il);

			ty->tlist = tnsav;
			allpushed += struct_lvalue->size;
		}	

		/* Hidden pointer is passed in first GPR! */
		ii = icode_make_addrof(struct_lvalue, il);
		append_icode_list(il, ii);
		free_preg(amd64_argregs[0], il, 1, 1);
		icode_make_copyreg(amd64_argregs[0], ii->dat, NULL, NULL, il);
		++regs_used;
	}

	if (fcall->functype->nargs == -1
		|| ty->implicit) {
		/* Need default argument promotions */
		need_dap = 1;
	}	

	for (i = 0; i < nvrs; ++i) {
		struct reg		*curreg;

		if (fcall->functype->variadic
			&& i >= fcall->functype->nargs) {
			need_dap = 1;
		}

		was_struct = was_float = was_llong = 0;
		if (vrs[i]->parent) {
			vr2 = get_parent_struct(vrs[i]);
		} else {
			vr2 = NULL;
		}

		if (vrs[i]->type->tlist != NULL
			|| is_integral_type(vrs[i]->type)) {
			if (regs_used < N_ARGREGS) {
				curreg = amd64_argregs[regs_used];
			} else {
				curreg = NULL;
			}

			if ((IS_CHAR(vrs[i]->type->code)
				|| IS_SHORT(vrs[i]->type->code))
				&& vrs[i]->type->tlist == NULL) {
				vrs[i] = backend->
					icode_make_cast(vrs[i],
						make_basic_type(TY_INT), il);
			} else {
				vreg_faultin(NULL, NULL, vrs[i], il, 0);
			}
			/*oldtype = vrs[i]->type;*/
			
			if (curreg != NULL) {
				struct reg	*topcurreg = curreg;

				if (curreg->size > vrs[i]->size) {
					curreg = curreg->composed_of[0];
				}
				if (vrs[i]->pregs[0] != curreg) {
					free_preg(topcurreg, il, 1, 1);
					icode_make_copyreg(curreg,
						vrs[i]->pregs[0],
						vrs[i]->type,
						vrs[i]->type, il);
				}
				reg_set_unallocatable(curreg);
				amd64_argregs[regs_used]->used = 0;
				++regs_used;
			} else {
				/* Pass remaining args on stack */
				passed_stack = 1;
				allpushed += pass_args_stack(vrs, i, nvrs, il);
				break;
			}
		} else if (IS_FLOATING(vrs[i]->type->code)) {
			if (vrs[i]->type->code == TY_LDOUBLE) {
				/* long double is always passed on stack */
				;
#if 0
				pass_ldouble_stack(vrs[i], &allpushed, il);
#endif
			} else {
				/* float or double */
				struct reg	*curfpreg;

				if (vrs[i]->type->code == TY_FLOAT
					&& need_dap) {
					struct type	*ty =
						make_basic_type(TY_DOUBLE);
					vrs[i] = backend->icode_make_cast(
						vrs[i], ty, il);
				}

				if (fp_regs_used < 8) {
					curfpreg = &amd64_sse_regs
						[fp_regs_used];
					if (vrs[i]->pregs[0] != curfpreg
						|| vrs[i]->pregs[0]->vreg
						!= vrs[i]) {
						free_preg(curfpreg,
							il, 1, 1);
					}
					vreg_faultin(curfpreg, NULL,
						vrs[i], il, 0);
					++fp_regs_used;
				} else {
					unimpl();
				}
			}
		} else if (vrs[i]->type->code == TY_STRUCT
				|| vrs[i]->type->code == TY_UNION) {
			pass_struct_union(vrs[i], &regs_used, &fp_regs_used,
				&allpushed, il);
		} else {
			unimpl();
		}
	}
	if (!passed_stack) {
		/* There may be long doubles left to pass */
		allpushed += pass_args_stack(vrs, i, nvrs, il);
	}

	/*
	 * In the x86 ABI, the caller is responsible for saving
	 * eax/ecx/edx (but not ebx, esi, edi), so that's what we 
	 * do here
	 */
	backend->invalidate_gprs(il, 1);
	if (fcall->functype->variadic || need_dap) {
		/* rax = number of sse registers used for call */
		ii = icode_make_setreg(&amd64_x86_gprs[0], fp_regs_used);
		append_icode_list(il, ii);
		reg_set_unallocatable(&amd64_x86_gprs[0]);
		amd64_x86_gprs[0].used = 0;
	}


	if (ty->tlist->type == TN_POINTER_TO) {
		/* Need to indirect thru function pointer */
		vreg_faultin(NULL, NULL, tmpvr, il, 0);
		ii = icode_make_call_indir(tmpvr->pregs[0]);
		tmpvr->pregs[0]->used = 0;
		tmpvr->pregs[0]->vreg = NULL;
	} else {
		ii = icode_make_call(ty->name);
	}	
	append_icode_list(il, ii);
	ii = icode_make_freestack(allpushed);
	append_icode_list(il, ii);

	for (i = 0; i < N_ARGREGS; ++i) {
		reg_set_allocatable(amd64_argregs[i]);
	}
	reg_set_allocatable(&amd64_x86_gprs[0]);

	ret = vreg_alloc(NULL, NULL, NULL, NULL);
	ret->type = ty;

	/* XXX man, this pointer stuff is painful and error prone */
	if ((ty->tlist->type == TN_POINTER_TO
		&& ty->tlist->next->next != NULL)
		|| (ty->tlist->type == TN_FUNCTION
		&& ty->tlist->next != NULL)) {
		/* Must be pointer */
		ret->pregs[0] = &amd64_x86_gprs[0];
	} else {
		if (IS_CHAR(ty->code)) {
			ret->pregs[0] = x86_gprs[0].composed_of[0]->
				composed_of[1];
		} else if (IS_SHORT(ty->code)) {
			ret->pregs[0] = x86_gprs[0].composed_of[0];
		} else if (IS_INT(ty->code)
			|| ty->code == TY_ENUM) { /* XXX */
			ret->pregs[0] = &x86_gprs[0];
		} else if (IS_LONG(ty->code) || IS_LLONG(ty->code)) {
			ret->pregs[0] = &amd64_x86_gprs[0];
		} else if (ty->code == TY_FLOAT
			|| ty->code == TY_DOUBLE) {
			ret->pregs[0] = &amd64_sse_regs[0];
		} else if (ty->code == TY_LDOUBLE) {
			ret->pregs[0] = &x86_fprs[0];
		} else if (ty->code == TY_STRUCT
			|| ty->code == TY_UNION) {
			ret->struct_ret = 1;
		} else if (ty->code == TY_VOID) {
			; /* Nothing! */
		}
	}

	ret->type = n_xmemdup(ret->type, sizeof *ret->type);
	if (ret->type->tlist->type == TN_POINTER_TO) {
		copy_tlist(&ret->type->tlist, ret->type->tlist->next->next);
	} else {
		copy_tlist(&ret->type->tlist, ret->type->tlist->next);
	}	
	if (ret->type->code != TY_VOID || ret->type->tlist) {
		ret->size = backend->get_sizeof_type(ret->type, NULL);
	}

	if (ret->pregs[0] != NULL) {
		vreg_map_preg(ret, ret->pregs[0]);
	}	
	
	if (is_x87_trash(ret)) {
		/*
		 * Don't keep stuff in x87 registers, ever!!
		 */
		free_preg(ret->pregs[0], il, 1, 1);
	}
	return ret;
}

static int
icode_make_return(struct vreg *vr, struct icode_list *il) {
	struct icode_instr	*ii;
	struct type_node	*oldtn;
	struct type		*rtype = curfunc->proto->dtype;

	oldtn = rtype->tlist;
	rtype->tlist = rtype->tlist->next;

	if (vr != NULL) {
		if (IS_CHAR(rtype->code)
			|| IS_SHORT(rtype->code)
			|| IS_INT(rtype->code)
			|| IS_LONG(rtype->code)
			|| IS_LLONG(rtype->code)
			|| rtype->tlist != NULL) {
			struct reg	*r = &amd64_x86_gprs[0];
			if (r->size > vr->size) {
				r = get_smaller_reg(r, vr->size);
			}
			vreg_faultin(r, NULL, vr, il, 0);
		} else if (rtype->code == TY_FLOAT
			|| rtype->code == TY_DOUBLE) {
			/* Return in xmm0 */
			vreg_faultin(&amd64_sse_regs[0], NULL, vr, il, 0);
		} else if (rtype->code == TY_LDOUBLE) {
			/* Return in st0 */
			vreg_faultin_x87(NULL, NULL, vr, il, 0);
		} else if (rtype->code == TY_STRUCT
			|| rtype->code == TY_UNION) {

#if 0
			struct decl		*la;
			struct stack_block	*sb;
			struct vreg		*dest;
			struct vreg		*from_ptr;
			static struct decl	dec;
			struct decl		*decp;
			unsigned long		offset;
			static struct type_node	tn;

			la = oldtn->tfunc->lastarg;

			/* Get hidden struct pointer for storing return */
			offset = 8; /* Move past ebp,eip */
			sb = make_stack_block(offset, 4);
			sb->is_func_arg = 1;
			dec.stack_addr = sb;
			dec.dtype = n_xmemdup(rtype, sizeof *rtype);
			decp = n_xmemdup(&dec, sizeof dec);
			from_ptr = vreg_alloc(decp, NULL, NULL, NULL);
			
			tn.type = TN_POINTER_TO;
			from_ptr->type->tlist = &tn;
			from_ptr->size = 4;
			vreg_faultin(NULL, NULL, from_ptr, il, 0);
			
			dest = vreg_alloc(NULL, NULL, NULL, NULL);
			dest->from_ptr = from_ptr;
#endif

			/* vr may come from pointer */
			vreg_faultin_ptr(vr, il);
			icode_make_copystruct(/*dest*/NULL, vr, il); 
		}
	}
	ii = icode_make_ret(vr);
	append_icode_list(il, ii);

	rtype->tlist = oldtn;

	return 0;
}

/*
 * Deal with preparations necessary to make things work with the terrible
 * x86 design
 */
static void 
icode_prepare_op(
	struct vreg **dest0,
	struct vreg **src0,
	int op,
	struct icode_list *il) {

	x86_backend.icode_prepare_op(dest0, src0, op, il);
}



/*
 * Most of the time, instructions give meaning to data. This function
 * generates code required to convert virtual register ``src'' to type
 * ``to'' where necessary
 */
static struct vreg *
icode_make_cast(struct vreg *src, struct type *to, struct icode_list *il) {
	return x86_backend.icode_make_cast(src, to, il);
}

static void
do_print_gpr(struct reg *r) {
	printf("%s=%d ", r->name, r->used);
	if (r->vreg && r->vreg->pregs[0] == r) {
		printf("<-> %p", r->vreg);
	}
}	

static void
debug_print_gprs(void) {
	int	i;
	
	for (i = 0; i < 6; ++i) {
		printf("\t");
		do_print_gpr(&amd64_x86_gprs[i]);
		printf("\t");
		do_print_gpr(&x86_gprs[i]);
		putchar('\t');
		do_print_gpr(x86_gprs[i].composed_of[0]);
		if (i < 4) {
			putchar('\t');
			do_print_gpr(x86_gprs[i].composed_of[0]->
				composed_of[0]);
			putchar('\t');
			do_print_gpr(x86_gprs[i].composed_of[0]->
				composed_of[1]);
		}	
		putchar('\n');
	}
}

static int
is_multi_reg_obj(struct type *t) {
	(void) t;
	return 0;
}	

struct backend amd64_backend = {
	ARCH_AMD64,
	0, /* ABI */
	0, /* multi_gpr_object */
	4, /* structure alignment */
	init,
	is_multi_reg_obj,
	get_ptr_size,
	get_size_t,
	get_sizeof_basic,
	get_sizeof_type,
	get_sizeof_elem_type,
	get_sizeof_decl,
	get_sizeof_const,
	get_align_type,
	gen_program,
	NULL,
	NULL,
	invalidate_gprs,
	/*invalidate_except*/NULL,
	alloc_gpr,
	/*alloc_16_or_32bit_noesiedi*/NULL,
	alloc_fpr,
	x86_free_preg,
	icode_make_fcall,
	icode_make_return,
	NULL,
	icode_prepare_op,
	icode_make_cast,
	NULL, /* icode_make_structreloc */
	make_null_block,
	make_init_name,
	debug_print_gprs,
	/*name_to_reg XXX */ NULL,
	/*asmvreg_to_reg*/ NULL,
	/*get_inlineasm_label*/NULL,
	do_ret
};

