/*
 * Copyright (c) 2006 - 2007, Nils R. Weller
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 * SGI assembler emitter
 */
#include "mips_emit_as.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <ctype.h>
#include "scope.h"
#include "type.h"
#include "decl.h"
#include "icode.h"
#include "subexpr.h"
#include "backend.h"
#include "token.h"
#include "functions.h"
#include "symlist.h"
#include "typemap.h"
#include "mips_gen.h"
#include "expr.h"
#include "reg.h"
#include "libnwcc.h"
#include "inlineasm.h"
#include "error.h"
#include "n_libc.h"

static FILE		*out;
static size_t		data_segment_offset;
static size_t		bss_segment_offset;
extern struct vreg	float_conv_mask;

static int 
init(FILE *fd, struct scope *s) {
	(void) s;
	out = fd;
	return 0;
}

static void
print_mem_operand(struct vreg *vr, struct token *constant,
struct vreg *vreg_parent);

static void
emit_addrof(struct reg *dest, struct vreg *src, struct vreg *stop);

static void
emit_load(struct reg *r, struct vreg *vr);

void	as_print_string_init(FILE *out, size_t howmany, struct ty_string *str);

static void
do_softarit_call(struct icode_instr *ii,
	const char *func, int op, int formod);


static void
do_emit_li(const char *dest, unsigned number, int has_sign) {
	int	need_neg = 0;

	if (has_sign && (int)number < 0) {
		number = -(int)number;
		need_neg = 1;
	}	
	/*
 	 * XXX cross-compilation :-( ...
 	 */
	if (number > 32767) {
		x_fprintf(out, "\tlis %s, %u\n",
			dest, number >> 16);
		x_fprintf(out, "\tori %s, %s, %u\n",
			dest, dest, number & 0xffff);
	} else {
		x_fprintf(out, "\tli %s, %u\n", dest, number);
	}
	if (need_neg) {
		x_fprintf(out, "\tneg %s, %s\n", dest, dest);
	}	
}

static void
print_init_expr(struct type *dt, struct expr *ex) {
	as_align_for_type(out, dt);
	x_fprintf(out, "\t");
	if (dt->tlist == NULL) {
		switch (dt->code) {
		case TY_CHAR:
		case TY_SCHAR:	
		case TY_UCHAR:
		case TY_BOOL:	
#if 0
			x_fprintf(out, ".byte 0x%x",
				*(unsigned char *)ex->
				const_value->value);
#endif
			x_fprintf(out, ".byte ");
			cross_print_value_by_type(out,
				ex->const_value->value, TY_UCHAR, 'x');	
			break;
		case TY_SHORT:
		case TY_USHORT:
#if 0
			x_fprintf(out, ".short 0x%x",
				*(unsigned short *)ex->
				const_value->value);
#endif
			x_fprintf(out, ".short ");
			cross_print_value_by_type(out,
				ex->const_value->value, TY_USHORT, 'x');	
			break;
		case TY_INT:
		case TY_ENUM:	
		case TY_UINT:
#if 0
			x_fprintf(out, ".long 0x%x",
				(unsigned)*(int *)ex->
				const_value->value);
#endif
			x_fprintf(out, ".long ");
			cross_print_value_by_type(out,
				ex->const_value->value, TY_UINT, 'x');	
			break;
#if 0
			x_fprintf(out, ".long 0x%x",
				*(unsigned int *)ex->
				const_value->value);
#endif
			break;
		case TY_LONG:
			if (backend->abi == ABI_POWER64) {
#if 0
				x_fprintf(out, ".llong 0x%lx",
					(unsigned long)*(long *)ex->
					const_value->value);
#endif
				x_fprintf(out, ".llong ");
				cross_print_value_by_type(out,
					ex->const_value->value, dt->code, 'x');
				
			} else {	
#if 0
				x_fprintf(out, ".long 0x%x",
					(unsigned)*(long *)ex->
					const_value->value);
#endif
				x_fprintf(out, ".long ");
				cross_print_value_by_type(out,
					ex->const_value->value, dt->code, 'x');
			}
			break;
		case TY_ULONG:
			if (backend->abi == ABI_POWER64) {
#if 0
				x_fprintf(out, ".llong 0x%x",
					(unsigned long)*(unsigned long *)ex->
					const_value->value);
#endif
				x_fprintf(out, ".llong ");
				cross_print_value_by_type(out,
					ex->const_value->value, dt->code, 'x');
			} else {	
#if 0
				x_fprintf(out, ".long 0x%x",
					(unsigned)*(unsigned long *)ex->
					const_value->value);
#endif
				x_fprintf(out, ".long ");
				cross_print_value_by_type(out,
					ex->const_value->value, dt->code, 'x');
			}	
			break;
		case TY_LLONG:
		case TY_ULLONG:	
#if 0
			x_fprintf(out, ".long 0x%x\n",
				*(unsigned int *)ex->
				const_value->value);
			x_fprintf(out, "\t.long 0x%x",
				((unsigned int *)ex->
				 const_value->value)[1]);
#endif
			x_fprintf(out, ".long ");
			cross_print_value_chunk(out,
				ex->const_value->value, dt->code,
				TY_UINT, 0, 0);
			x_fputc('\n', out);
			x_fprintf(out, "\t.long ");
			cross_print_value_chunk(out,
				ex->const_value->value, dt->code,
				TY_UINT, 0, 1);

			break;
		case TY_FLOAT:
#if 0
			x_fprintf(out, ".long 0x%x\n",
				*(unsigned int *)ex->
				const_value->value);
#endif
			x_fprintf(out, ".long ");
			cross_print_value_by_type(out,
				ex->const_value->value, TY_UINT, 'x');	
			break;
		case TY_DOUBLE:
		case TY_LDOUBLE:
#if 0
			x_fprintf(out, ".long 0x%x\n",
				*(unsigned int *)ex->
				const_value->value);
			x_fprintf(out, "\t.long 0x%x",
				((unsigned int *)ex->
				 const_value->value)[1]);
#endif
			x_fprintf(out, ".long ");
			cross_print_value_chunk(out,
				ex->const_value->value, dt->code,
				TY_UINT, 0, 0);
			x_fputc('\n', out);
			x_fprintf(out, "\t.long ");
			cross_print_value_chunk(out,
				ex->const_value->value, dt->code,
				TY_UINT, 0, 1);
			break;
		default:	
			printf("print_init_expr: "
				"unsupported datatype %d\n",
				dt->code);
			unimpl();
		}
	} else {
		if (dt->tlist->type == TN_POINTER_TO) {
			struct tyval	*cv;

			cv = ex->const_value;
			if (backend->abi == ABI_POWER64) {
				x_fprintf(out, ".llong ");
			} else {
				x_fprintf(out, ".long ");
			}	
			if (cv->is_nullptr_const) {
				x_fprintf(out, "0x0");
			} else if (cv->str) {
				x_fprintf(out, "_Str%lu_data", cv->str->count);
			} else if (cv->value) {
#if 0
				x_fprintf(out, "%lu",
					*(unsigned long *)cv->value);
#endif
				cross_print_value_by_type(out,
					cv->value, TY_ULONG, 'd');	
			} else if (cv->address) {
				/* XXX */
				char	*sign;

				if (cv->address->diff < 0) {
					sign = "-";
				} else {
					sign = "+";
				}
				x_fprintf(out, "%s%s%ld",
					cv->address->dec->dtype->name, sign,
					cv->address->diff);
			}	
		} else if (dt->tlist->type == TN_ARRAY_OF) {
			size_t	arrsize;
			struct tyval	*cv;
			
			/*
			 * This has to be a string because only in
			 * char buf[] = "hello"; will an aggregate
			 * initializer ever be stored as INIT_EXPR
			 */

			cv = ex->const_value;
			arrsize = dt->tlist->arrarg_const;
			as_print_string_init(out, arrsize, cv->str);

			if (arrsize >= cv->str->size) {
				if (arrsize > cv->str->size) {
					x_fprintf(out, "\n\t.space %lu\n",
						arrsize - cv->str->size);
				}
			} else {
				/* Do not null-terminate */
				;
			}
		} else { /* function */
			struct tyval	*cv = ex->const_value;
			x_fprintf(out, ".%s %s",
				backend->abi == ABI_POWER64? "llong": "long",
				cv->address->dec->dtype->name);
		}
	}
	x_fputc('\n', out);
}


static void
emit_extern_decls(void) {
	struct sym_entry	*se;

	for (se = extern_vars; se != NULL; se = se->next) {
		if (se->dec->has_symbol
			|| se->dec->dtype->is_def
			|| se->dec->has_def) {
			continue;
		}
		if (se->dec->references == 0
			&& (!se->dec->dtype->is_func
				|| !se->dec->dtype->is_def)) {
			/* Unneeded declaration */
			continue;
#if 0
		} else if (se->dec->dtype->is_func
			&& !se->dec->dtype->is_def) {
			continue;
#endif
		}
		x_fprintf(out, "_Toc_%s:\n", se->dec->dtype->name);
		x_fprintf(out, "\t.tc %s[TC], %s[%s]\n",
			se->dec->dtype->name, se->dec->dtype->name,
				se->dec->dtype->is_func
				&& !se->dec->dtype->is_def?
				"DS": "RW");
	}
}


static void 
emit_static_decls(void) {
	struct decl	*d;
	struct decl	**dv;
	struct function	*func;
	size_t		size;
	int		i;

	(void) d; (void) size;

	bss_segment_offset = 0;
	x_fprintf(out, "\t.lcomm _Divbuf0_data, 8, _%s.bss_\n",
		tunit_name);
	x_fprintf(out, "\t.lcomm _Divbuf1_data, 8, _%s.bss_\n",
		tunit_name);
	if (static_uninit_vars != NULL) {
		for (d = static_uninit_vars; d != NULL; d = d->next) {
			if (d->dtype->storage == TOK_KEY_STATIC
				&& d->references == 0) {
				continue;
			}

			as_align_for_type(out, d->dtype);
			size = backend->get_sizeof_decl(d, NULL);
			x_fprintf(out, "\t.%s ",
				d->dtype->storage == TOK_KEY_STATIC?
				"lcomm": "comm");
			x_fprintf(out, "%s, %d, _%s.bss_\n",
				d->dtype->name, size, tunit_name);
		}
	}

	data_segment_offset = 0;
	if (static_init_vars != NULL) {
		x_fprintf(out, "\t.csect .data[RW], 3\n");
		x_fprintf(out, "\t.align 2\n");
	}
	for (d = static_init_vars; d != NULL; d = d->next) {
		data_segment_offset += generic_print_init_var(out,
			d, data_segment_offset, print_init_expr, 1);
	}

	x_fprintf(out, "\t.toc\n");
	x_fprintf(out, "_Divbuf0:\n");
	x_fprintf(out, "\t.tc _Divbuf0_data[TC], _Divbuf0_data\n");
	x_fprintf(out, "_Divbuf1:\n");
	x_fprintf(out, "\t.tc _Divbuf1_data[TC], _Divbuf1_data\n");
	for (d = static_uninit_vars; d != NULL; d = d->next) {
		x_fprintf(out, "_Toc_%s:\n", d->dtype->name);
		x_fprintf(out, "\t.tc %s[TC], %s\n",
			d->dtype->name, d->dtype->name);
	}

	for (func = funclist; func != NULL; func = func->next) {
		char	*fname = func->proto->dtype->name;

		x_fprintf(out, "_Toc_%s:\n", fname);
		x_fprintf(out, "\t.tc %s[TC], %s\n",
			fname, fname);
	}
	for (d = static_init_vars; d != NULL; d = d->next) {
		if (d->is_alias) continue;
		x_fprintf(out, "_Toc_%s:\n", d->dtype->name);
		x_fprintf(out, "\t.tc %s[TC], %s\n",
			d->dtype->name, d->dtype->name);
	}

	x_fprintf(out, "\t.csect .text[PR]\n");
	dv = global_scope.static_decls.data;
	for (i = 0; i < global_scope.static_decls.ndecls; ++i) {
		if (dv[i]->dtype->storage != TOK_KEY_STATIC) {
			struct type_node	*tn = NULL;
			if (dv[i]->dtype->is_func) {
				for (tn = dv[i]->dtype->tlist;
					tn != NULL;
					tn = tn->next) {
					if (tn->type == TN_FUNCTION) {
						if (tn->ptrarg) {
							continue;
						} else {
							break;
						}	
					}
				}
			}
			x_fprintf(out, ".globl %s\n",
				dv[i]->dtype->name);
			if (tn != NULL) {
				tn->ptrarg = 1;
			}	
		}
	}
}

static void
emit_struct_inits(void) {
	struct init_with_name	*in;

	x_fprintf(out, "\t.csect .data[RW]\n");
	for (in = init_list_head; in != NULL; in = in->next) {
		as_align_for_type(out, make_basic_type(TY_LONG)); /* XXX */
		x_fprintf(out, "%s_data:\n", in->name);
		generic_print_init_list(out, in->dec,
			in->init, print_init_expr);
	}
	x_fprintf(out, "\t.toc\n");
	for (in = init_list_head; in != NULL; in = in->next) {
		x_fprintf(out, "%s:\n", in->name);
		x_fprintf(out, "\t.tc %s_data[TC], %s_data\n",
			 in->name, in->name);
	}
}

/*
 * XXX Misleading name ;o
 */
static void
emit_strings(void) {
	struct ty_string	*str;
	struct ty_float		*tf;
	struct ty_llong		*tll;

	if (str_const == NULL
		&& float_const == NULL
		&& llong_const == NULL) {
		return;
	}	

	x_fprintf(out, "\t.csect text[PR]\n");
	if (str_const != NULL) {
		for (str = str_const; str != NULL; str = str->next) {
			x_fprintf(out, "_Str%lu_data:\n", str->count);
			as_print_string_init(out, str->size, str);
		}
	}

	if (unimpl_instr) {
		x_fprintf(out,"\t_Unimpl_msg db 'ERROR: Use of unimplemented'\n"
			"\t            db 'compiler feature (probably)'\n"
			"\t            db 'floating point - check %d\n'\n");
	}	
	x_fprintf(out, "\t.align 3\n");

	if (float_const != NULL
		|| floatbuf.var_backed != NULL) {
	}

	if (float_const != NULL
		|| float_conv_mask.var_backed
		|| llong_const != NULL) {
		x_fprintf(out, "\t.toc\n");
		if (float_conv_mask.var_backed != NULL) {
			x_fprintf(out, "%s:\n",
				 float_conv_mask.var_backed->dtype->name);
			x_fprintf(out, "\t.tc %s_hmm[TC], ",
				float_conv_mask.var_backed->dtype->name);
			x_fprintf(out, "0x43300000, 0x80000000\n");
		}
		for (tf = float_const; tf != NULL; tf = tf->next) {
			/* XXX cross-compilation ... */
			x_fprintf(out, "_Float%lu:\n", tf->count);
			x_fprintf(out, "\t.tc _Float%lu_",
				tf->count);
			cross_print_value_by_type(out, tf->num->value,
				TY_UINT, 'x');
			x_fprintf(out, "[TC], ");

			if (backend->abi != ABI_POWER64) {
				cross_print_value_by_type(out, tf->num->value,
					TY_UINT, 'x');
				if (tf->num->type == TY_DOUBLE
					|| tf->num->type == TY_LDOUBLE) {
					x_fprintf(out, ", ");
					cross_print_value_chunk(out,
						tf->num->value,
						TY_DOUBLE, TY_UINT, 0, 1);
				} 
			} else {
				/*
				 * That
				 *    .tc blabla[TC], 0xstuff, 0xstuff
				 * stuff doesn't work in 64bit mode,
				 * because those two numbers are 64bit
				 * now instead of 32bit.
				 * So we print doubles/long doubles as
				 * a single 64bit number
				 */
				if (tf->num->type == TY_FLOAT) {
					cross_print_value_by_type(out,
						tf->num->value, TY_UINT, 'x');

					/*
					 * The number we just printed is a
					 * 32bit value, which without
					 * intervention would become the
					 * low word of a double word. But
					 * we want it to become the high
					 * one, so print 32 zero bits as low
					 * part!
					 */
					x_fprintf(out, "00000000");
				} else {
					cross_print_value_by_type(out,
						tf->num->value, TY_ULLONG, 'x');
				}
			}

			x_fputc('\n', out);
		}
		if (llong_const != NULL) {
			for (tll = llong_const; tll != NULL; tll = tll->next) {
#if 0
				/* this breaks for some reason :/ */
				if (!tll->loaded) {
					/* Not actually loaded */
					continue;
				}	
#endif
				x_fprintf(out, "_Llong%lu:\n",
					tll->count);
#if 0
				x_fprintf(out, "\t.tc _Llong%lu_%llx"
					"[TC], 0x%llx\n",
					/* XXX c99/gnu :-( */
					tll->count,
					*(unsigned long long *)tll->num->value,
					*(unsigned long long *)tll->num->value);
#endif
				x_fprintf(out, "\t.tc _Llong%lu_", tll->count);
				cross_print_value_by_type(out, tll->num->value,
					TY_ULLONG, 'x');
				x_fprintf(out, "[TC], ");
				cross_print_value_by_type(out, tll->num->value,
					TY_ULLONG, 'x');
				x_fputc('\n', out);
			}
		}
	}

	x_fprintf(out, ".toc\n");
	for (str = str_const; str != NULL; str = str->next) {
		x_fprintf(out, "_Str%lu:\n", str->count);
		x_fprintf(out, "\t.tc _Str%lu_data[TC], _Str%lu_data\n",
			str->count, str->count);
	}

	if (floatbuf.var_backed != NULL) {
	}
}	

static void
emit_comment(const char *fmt, ...) {
	int		rc;
	va_list	va;

	va_start(va, fmt);
	x_fprintf(out, "; ");
	rc = vfprintf(out, fmt, va);
	va_end(va);
	x_fputc('\n', out);

	if (rc == EOF || fflush(out) == EOF) {
		perror("vfprintf");
		exit(EXIT_FAILURE);
	}
}

static void
emit_inlineasm(struct inline_asm_stmt *stmt) {
	unimpl();
	x_fprintf(out, "; inline start\n");
	/*
	 * There may be an empty body for statements where only the side
	 * effect is desired;
	 * __asm__("" ::: "memory");
	 */
	if (stmt->code != NULL) {
		inline_instr_to_nasm(out, stmt->code);
	}	
	x_fprintf(out, "; inline end\n");
}	

static void
emit_unimpl(void) {
	/* Not implemented!! HA HA HA HOW IRONIC!!!!!!!!!!!!!! */
	unimpl();
}	

static void
emit_empty(void) {
	x_fputc('\n', out);
}

static void
emit_label(const char *name, int is_func) {
	if (is_func) {
		x_fprintf(out, "%s:\n", name);
	} else {
		x_fprintf(out, ".%s:\n", name);
	}
}

static void
emit_call(const char *name) {
	x_fprintf(out, "\tbl .%s\n", name);
	x_fprintf(out, "\tnop\n");
}

static void
emit_callindir(struct reg *r) {
	if (backend->abi == ABI_POWER64) {
		x_fprintf(out, "\tstd 2, 40(1)\n");
		x_fprintf(out, "\tld 0, 0(%s)\n", r->name);
		x_fprintf(out, "\tmtctr 0\n");
		x_fprintf(out, "\tld 2, 8(%s)\n", r->name); 
		x_fprintf(out, "\tld 11, 16(%s)\n", r->name);
		x_fprintf(out, "\tbctrl\n");
		x_fprintf(out, "\tnop\n");
		x_fprintf(out, "\tld 2, 40(1)\n");
	} else {	
		x_fprintf(out, "\tstw 2, 20(1)\n");
		x_fprintf(out, "\tlwz 0, 0(%s)\n", r->name);
		x_fprintf(out, "\tmtctr 0\n");
		x_fprintf(out, "\tlwz 2, 4(%s)\n", r->name); 
		x_fprintf(out, "\tlwz 11, 8(%s)\n", r->name);
		x_fprintf(out, "\tbctrl\n");
		x_fprintf(out, "\tnop\n");
		x_fprintf(out, "\tlwz 2, 20(1)\n");
	}	
}	

static void
emit_func_intro(struct function *f) {
	(void) f;
}

static void
emit_func_outro(void) {
}


static void
emit_allocstack(struct function *f, size_t nbytes) {
	(void) f;
	if (!f->gotframe) {
		/* Creating stack frame */
		if (nbytes > 32767) {
			/*
			 * More than 16 bits - need stwux
			 */
			unsigned int	ui = (unsigned)-nbytes;

			/* XXX 64bit. XXX cross-comp */
#if 0
			x_fprintf(out, "\tlis 0, 0x%x\n",
				((unsigned short *)&ui)[0]);
			x_fprintf(out, "\tori 0, 0, 0x%x\n",
				((unsigned short *)&ui)[1]);
#endif
			x_fprintf(out, "\tlis 0, ");
			cross_print_value_chunk(out, &ui, TY_UINT, TY_USHORT,
				0, 0);	
			x_fprintf(out, "\n\tori 0, 0, ");
			cross_print_value_chunk(out, &ui, TY_UINT, TY_USHORT,
				0, 1);
			x_fputc('\n', out);
			x_fprintf(out, "\t%s 1, 1, 0\n",
				backend->abi == ABI_POWER64? "stdux": "stwux");
		} else {
			x_fprintf(out, "\t%s 1, -%lu(1)\n",
				backend->abi == ABI_POWER64? "stdu": "stwu",
				(unsigned long)nbytes);	
			x_fprintf(out, "\tmr 31, 1\n");
		}
		f->gotframe = 1;
	} else {
		x_fprintf(out, "\tsi 1, 1, %lu\n",
			(unsigned long)nbytes);
	}
}

/*
 * Never use addi/subi! Those can't take a gpr0 source argument 
 */
static void
do_addi(struct reg *dest, struct reg *src, int val) {
	x_fprintf(out, "\taddic %s, %s, %d\n", dest->name, src->name, val);
}

static void
emit_freestack(struct function *f, size_t *nbytes) {
	char	*instr = "addic";

	if (nbytes == NULL) {
		/* Procedure outro */
		if (f->total_allocated != 0) {
			/* XXX never needed? */
#if 0
			x_fprintf(out, "\t%s 1, 1, %lu\n",
				instr, (unsigned long)f->total_allocated);
#endif
			x_fprintf(out, "\t%s 1, 0(1)\n",
				backend->abi == ABI_POWER64? "ld": "lwz");
		}	
	} else {
		if (*nbytes != 0) {
			x_fprintf(out, "\t%s 1, 1, %lu\n",
				instr, (unsigned long)*nbytes);	
			f->total_allocated -= *nbytes;
		}
	}
}

static void
emit_adj_allocated(struct function *f, int *nbytes) {
	f->total_allocated += *nbytes; 
}	

static void
emit_struct_defs(void) {
	struct scope	*s;

	for (s = &global_scope; s != NULL; s = s->next) {
		struct ty_struct	*ts;

		for (ts = s->struct_defs.head; ts != NULL; ts = ts->next) {
			if (!ts->is_union) {
				/*do_print_struct(ts);*/
			}	
		}
	}
}

static int	cursect = 0;

static void
emit_setsect(int value) {
	cursect = value;
}

static void
emit_alloc(size_t nbytes) {
	(void) nbytes;
	unimpl();
}


static void
print_mem_or_reg(struct reg *r, struct vreg *vr) {
	if (vr->on_var) {
		unimpl();
	} else {
		x_fprintf(out, "%s", r->name);
	}
}



static void
emit_inc(struct icode_instr *ii) {
	if (ii->src_vreg->is_multi_reg_obj) {
		x_fprintf(out, "\taddic %s, %s, 1\n",
			ii->src_pregs[1]->name,
			ii->src_pregs[1]->name);
		x_fprintf(out, "\taddze %s, %s\n",
			ii->src_pregs[0]->name,
			ii->src_pregs[0]->name);
	} else {
		do_addi(ii->src_pregs[0], ii->src_pregs[0], 1);
	}
}

static void
emit_dec(struct icode_instr *ii) {
	if (ii->src_vreg->is_multi_reg_obj) {
		x_fprintf(out, "\tsubic %s, %s, 1\n",
			ii->src_pregs[1]->name,
			ii->src_pregs[1]->name);
		x_fprintf(out, "\taddme %s, %s\n",
			ii->src_pregs[0]->name,
			ii->src_pregs[0]->name);
	} else {
		x_fprintf(out, "\t%s %s, %s, 1\n",
			ii->src_vreg->size == 8? "subic": "subic",
			ii->src_pregs[0]->name, ii->src_pregs[0]->name);
	}
}


static void
emit_load(struct reg *r, struct vreg *vr) {
	char		*p = NULL;
	int		need_ext = 0;
	int		need_frsp = 0;
	int		static_var = 0;
	static int	was_llong;
#define EXT_RLWINM	1
#define EXT_EXTSH	2
	struct vreg	*vr2 = NULL;

	if (vr == &float_conv_mask) {
		/* XXX what a terrible kludge */
		x_fprintf(out, "\tlfd %s, %s(2)\n",
			 r->name,
			 float_conv_mask.var_backed->dtype->name);
		return;
	}

	/*
	 * If we need to do displacement addressing with a static
	 * variable, its address must be moved to a GPR first...
	 */
	if (vr->parent) { 
		vr2 = get_parent_struct(vr);
		if (vr2->stack_addr == NULL
			&& vr2->var_backed
			&& vr2->var_backed->dtype->storage
				!= TOK_KEY_REGISTER) {
			emit_addrof(tmpgpr, vr, vr2);
			static_var = 1;
		}
	}

	if (vr->var_backed
		&& vr->var_backed->stack_addr == NULL
		&& vr->parent == NULL
		&& (vr->type->tlist == NULL
		|| (vr->type->tlist->type != TN_ARRAY_OF
			&& vr->type->tlist->type != TN_FUNCTION))) {
		static_var = 1;
	}

	if (r->type == REG_FPR) {
		/* No conversion between int/fp is handled here! */
		if (vr->size == 4) {
			/* p = "l.s"; */
			p = "lfs";
			need_frsp = 1;
		} else if (vr->size == 8) {
			/* p = "l.d"; */
			p = "lfd";
		} else if (vr->size == 16) {
			/* long double */
			unimpl();
		} else {
			/* ?!?!? */
			unimpl();
		}
	} else {
		if (vr->type
			&& vr->type->tlist != NULL
			&& vr->type->tlist->type == TN_ARRAY_OF) {
			if (vr->from_const != NULL) { 
				p = backend->abi == ABI_POWER64? "ld": "lwz";
			} else {
				emit_addrof(r, vr, vr2);
				return; 
			}
		} else if (vr->from_const) {
			if (IS_LLONG(vr->type->code)
				&& backend->abi != ABI_POWER64) {
				do_emit_li(r->name,
					((unsigned *)
					vr->from_const->data)[was_llong],
					0);
				if (was_llong) {
					was_llong = 0;
				} else if (vr->is_multi_reg_obj) {
					was_llong = 1;
				}
				return;
			} else if (IS_LLONG(vr->type->code)
				|| (IS_LONG(vr->type->code)
					&& backend->abi == ABI_POWER64)) {
				x_fprintf(out, "\tld %s, ", r->name);
				print_mem_operand(vr, NULL, vr2);
				x_fputc('\n', out);
				return;
			} else {
				/*p = "li";*/
				/* XXXXXXXXXXXXXXX cross-compilation :-( */
				do_emit_li(r->name,
					*(unsigned *)vr->from_const->data,
					vr->type->sign != TOK_KEY_UNSIGNED);
				return;
			}
		} else {
			if (vr->type == NULL) {
				if (vr->size == 8) {
					if (backend->abi == ABI_POWER64) {
						p = "ld";
					} else {
						p = "lwz";
					}
				} else {
					p = "lwz";
				}
			} else if (vr->type->tlist != NULL) {	
				if (backend->abi == ABI_POWER64) {
					p = "ld";
				} else {
					p = "lwz";
				}
			} else {
				if (IS_CHAR(vr->type->code)) {
					p = "lbz";
					if (vr->type->code == TY_SCHAR) {
						need_ext = EXT_RLWINM;
					}
				} else if (vr->type->code == TY_SHORT) {
					/*p = "lha";*/
					p = "lhz";
					need_ext = EXT_EXTSH;
				} else if (vr->type->code == TY_USHORT) {
					p = "lhz";
				} else if (vr->type->code == TY_INT
					|| vr->type->code == TY_ENUM) {
					/* p = "lwa"; */
					p = "lwz";
				} else if (vr->type->code == TY_UINT) {
					p = "lwz";
				} else if (vr->type->code == TY_LONG) {
					if (backend->abi == ABI_POWER64) {
						p = "ld";
					} else {
						p = "lwz";
					}
				} else if (vr->type->code == TY_ULONG) {
					if (backend->abi == ABI_POWER64) {
						p = "ld";
					} else {
						p = "lwz";
					}
				} else if (vr->type->code == TY_LLONG) {
					if (backend->abi == ABI_POWER64) {
						p = "ld";
					} else {
						p = "lwz";
					}
				} else if (vr->type->code == TY_ULLONG) {
					if (backend->abi == ABI_POWER64) {
						p = "ld";
					} else {
						p = "lwz";
					}
				} else if (vr->type->code == TY_FLOAT) {
					p = "lw";
				} else if (vr->type->code == TY_DOUBLE) {
					p = "ld";
				} else if (vr->type->code == TY_LDOUBLE) {
					unimpl();
				} else {
					printf("cannot load %d\n",
						vr->type->code);
					unimpl();
				}
			}	
		}	
	}

	/*
	 * For static structs and nonstatic scalars, this loads the
	 * scalar itself. For static scalars, this loads the TOC
	 * pointer
	 * XXX whew this is VERY confusing
	 */
	if (backend->abi == ABI_POWER64) {
		x_fprintf(out, "\t%s %s, ",
			static_var && vr->parent == NULL? "ld": p,
			need_ext? power_gprs[0].name: r->name);
	} else {
		x_fprintf(out, "\t%s %s, ",
			static_var && vr->parent == NULL? "lwz": p,
			need_ext? power_gprs[0].name: r->name);
	}

	print_mem_operand(vr, NULL, vr2);
	x_fputc('\n', out);
	if (static_var && vr->parent == NULL) {
		/*
		 * Fetching scalar through TOC pointer in r
		 */
		x_fprintf(out, "\t%s %s, %s(%s)\n",
			p, /*r->name,*/
			need_ext? power_gprs[0].name: r->name,
			was_llong? "4": "0",
			r->name); 
		if (vr->is_multi_reg_obj) {
			if (was_llong) {
				was_llong= 0;
			} else {
				was_llong = 1;
			}
		}
	}

	if (need_ext) {
		if (need_ext == EXT_EXTSH) {
			x_fprintf(out, "\textsh %s, %s\n",
				power_gprs[0].name, power_gprs[0].name);
		} else {
			/* rlwinm */
			x_fprintf(out, "\trlwinm %s, %s, %s, 0xff\n",
				power_gprs[0].name, power_gprs[0].name,
				power_gprs[0].name);
		}
		x_fprintf(out, "\tmr %s, %s\n", r->name, power_gprs[0].name);
	}
}


/*
 * Takes vreg source arg - not preg - so that it can be either a preg
 * or immediate (where that makes sense!)
 */
static void
emit_store(struct vreg *dest, struct vreg *src) {
	char		*p = NULL;
	int		floating = 0;
	int		is_static = 0;
	static int	was_llong;
	struct vreg	*vr2 = NULL;

	if (dest->var_backed
		&& dest->parent == NULL
		&& dest->var_backed->stack_addr == NULL) {
		is_static = 1;
		emit_addrof(tmpgpr, dest, NULL);
	}

	if (dest->parent) {
		vr2 = get_parent_struct(dest);
		if (vr2->stack_addr == NULL
			&& vr2->var_backed
			&& vr2->var_backed->dtype->storage
			!= TOK_KEY_REGISTER) {
			emit_addrof(tmpgpr, dest, vr2);
			is_static = 1;
		}
	}

	if (src->pregs[0] && src->pregs[0]->type == REG_FPR) {	
		/* Performs no fp-int conversion! */
		if (src->type->code == TY_FLOAT) {
			p = "stfs";
			x_fprintf(out, "\tfrsp %s, %s\n",
				src->pregs[0]->name, src->pregs[0]->name);
		} else if (src->type->code == TY_DOUBLE) {
			p = "stfd";
		} else {
			unimpl();
		}

		floating = 1;
	} else {
		if (dest->type == NULL) {
			if (backend->abi == ABI_POWER64) {
				p = "std";
			} else {
				p = "stw";
			}	
		} else if (dest->type->tlist != NULL) {
			/* Must be pointer */
			if (backend->abi == ABI_POWER64) {
				p = "std";
			} else {
				p = "stw";
			}
		} else {
			if (IS_CHAR(dest->type->code)) {
				p = "stb";
			} else if (IS_SHORT(dest->type->code)) {
				p = "sth";
			} else if (IS_INT(dest->type->code)
				|| dest->type->code == TY_ENUM) {
				p = "stw";
			} else if (IS_LONG(dest->type->code)) {
				if (backend->abi == ABI_POWER64) {
					p = "std";
				} else {
					p = "stw";
				}
			} else if (IS_LLONG(dest->type->code)) {
				if (backend->abi == ABI_POWER64) {
					p = "std";
				} else {
					p = "stw";
				}
			} else {
			printf("well ... %d(%p)\n",
				dest->type->code, dest);
			printf("%s\n", src->pregs[0]->name); 
			printf("%d\n", src->pregs[0]->used);
			*(char *)0 = 0;
				unimpl();
			}	
		}	
	}

	x_fprintf(out, "\t%s ", p);
	if (src->from_const) {
		print_mem_operand(src, NULL, NULL);
		x_fprintf(out, ", ");
	} else {
		/* Must be register */
		if (was_llong) {
			x_fprintf(out, "%s, ", src->pregs[1]->name);
			was_llong = 0;
		} else {
			x_fprintf(out, "%s, ", src->pregs[0]->name);
			if (src->is_multi_reg_obj) {
				was_llong = 1;
			}
		}
	}
	if (!is_static) {
		print_mem_operand(dest, NULL, vr2);
	} else {
		x_fprintf(out, "0(%s)", tmpgpr->name);
	}
	x_fputc('\n', out);
}


static void
emit_neg(struct reg **dest, struct icode_instr *src) {
	(void) src;
	if (dest[0]->type == REG_FPR) {
		x_fprintf(out, "\tfneg %s, %s\n",
			dest[0]->name, dest[0]->name);
	} else {
		x_fprintf(out, "\tneg %s, %s\n",
			dest[0]->name, dest[0]->name);
	}	
}	


static void
emit_sub(struct reg **dest, struct icode_instr *src) {
	char	*p = NULL;

	if (dest[0]->type == REG_FPR) {
		if (src->src_vreg->type->code == TY_FLOAT) {
			p = "fsubs"; /*"fsub";*/
		} else if (src->src_vreg->type->code == TY_DOUBLE) {
			p = "fsub";
		} else {
			unimpl();
		}
		x_fprintf(out, "\t%s %s, %s, ",
			p, dest[0]->name, dest[0]->name); 
	} else {
		/* XXX take immediate into account */
		if (src->src_vreg->size == 8) {
			if (backend->abi != ABI_POWER64) {
				do_softarit_call(src, "__nwcc_sub",
					TOK_OP_MINUS, 0);
				return;
			} else {
				/*unimpl();*/
				p = "subf";
			}
		} else {
			/* XXX */
			p = "subfc";
		}
		
		x_fprintf(out, "\t%s %s, %s, %s\n",
			p, dest[0]->name,
			src->src_pregs[0]->name, dest[0]->name);
		return;
	}
	print_mem_or_reg(src->src_pregs[0], src->src_vreg);
	x_fputc('\n', out);
}

static void
emit_add(struct reg **dest, struct icode_instr *src) {
	char	*p = NULL;

	if (dest[0]->type == REG_FPR) {
		if (src->src_vreg->type->code == TY_FLOAT) {
			p = "fadds"; /*"fadd";*/
		} else if (src->src_vreg->type->code == TY_DOUBLE) {
			p = "fadd";
		} else {
			unimpl();
		}
		x_fprintf(out, "\t%s %s, %s, ",
			p, dest[0]->name, dest[0]->name);
	} else {
		/* XXX take immediate into account ... addi, etc */
		if (src->src_vreg->size == 8) {
			if (backend->abi != ABI_POWER64) {
				do_softarit_call(src, "__nwcc_add",
					TOK_OP_PLUS, 0);
				return;
			} else {
				/*unimpl();*/
				p = "add";
			}
		} else {
			p = "add";
		}
		x_fprintf(out, "\t%s %s, %s, ",
			p, dest[0]->name, dest[0]->name);
	}	
	print_mem_or_reg(src->src_pregs[0], src->src_vreg);
	x_fputc('\n', out);
}

/* Note that this is 32bit-only */
static void
do_softarit_call(struct icode_instr *src,
	const char *func, int op, int formod) {

	struct reg	**dest = src->dest_pregs;
	struct type	*ty = src->dest_vreg->type;

	/* Store data to div buffer 1 */
	x_fprintf(out, "\tlwz %s, _Divbuf0(2)\n", tmpgpr->name);
	x_fprintf(out, "\tstw %s, 0(%s)\n",
		dest[0]->name, tmpgpr->name);
	if (src->dest_vreg->is_multi_reg_obj) {
		x_fprintf(out, "\tstw %s, 4(%s)\n",
			dest[1]->name, tmpgpr->name);
	}

	if (op == TOK_OP_BSHL || op == TOK_OP_BSHR) {
		/* data, shiftbits, datasize, wrong_endianness */

		/* Load data buffer addresses */
		if (src->src_pregs[0] != &power_gprs[4]) {
			x_fprintf(out, "\tmr 4, %s\n",
				src->src_pregs[0]->name);
		}
		x_fprintf(out, "\tlwz 3, _Divbuf0(2)\n");
		x_fprintf(out, "\tli 5, %d\n",
			backend->get_sizeof_type(ty, NULL) * 8);
		x_fprintf(out, "\tli 6, 1\n");
		goto do_call;
	}

	/* Store data to div buffer 2 */
	x_fprintf(out, "\tlwz %s, _Divbuf1(2)\n", tmpgpr->name);
	x_fprintf(out, "\tstw %s, 0(%s)\n",
		src->src_pregs[0]->name, tmpgpr->name);
	if (src->src_vreg->is_multi_reg_obj) {
		x_fprintf(out, "\tstw %s, 4(%s)\n",
			src->src_pregs[1]->name, tmpgpr->name);
	}

	/* Load buffer addresses */
	x_fprintf(out, "\tlwz 3, _Divbuf0(2)\n");
	x_fprintf(out, "\tmr 4, %s\n", tmpgpr->name); 

	if (op == TOK_OP_DIVIDE) {
		/* Set want_remainder */
		x_fprintf(out, "\tli 5, %d\n", formod? 1: 0);
	}

	/* Set number of value bits */
	do_emit_li(op == TOK_OP_DIVIDE? "6": "5",
		backend->get_sizeof_type(ty, NULL) * 8, 0);
	if (op == TOK_OP_PLUS || op == TOK_OP_MINUS) {
		/* Has wrong endianness */
		do_emit_li("6", 1, 0);
		if (op == TOK_OP_MINUS) {
			/* maxbit = datasize.. maxbit unneeded?!? */
			do_emit_li("7", backend->get_sizeof_type(ty, NULL) * 8,
				0);
		}
	}

do_call:
	x_fprintf(out, "\tbl .%s\n", func);
	x_fprintf(out, "\tnop\n");

	/* Copy result from _Divbuf0 to result register */
	x_fprintf(out, "\tlwz %s, _Divbuf0(2)\n", tmpgpr->name);
	x_fprintf(out, "\tlwz %s, 0(%s)\n", dest[0]->name, tmpgpr->name);
	if (src->dest_vreg->is_multi_reg_obj) {
		x_fprintf(out, "\tlwz %s, 4(%s)\n", dest[1]->name,
			tmpgpr->name);
	}
}

static void
emit_div(struct reg **dest, struct icode_instr *src, int formod) {
	struct type	*ty = src->src_vreg->type;
	char		*p = NULL;

	(void) dest;

	if (IS_FLOATING(ty->code)) {
		if (ty->code == TY_FLOAT) {
			p = "fdivs"; /* "fdiv";*/
		} else if (ty->code == TY_DOUBLE) {
			p = "fdiv";
		} else {
			unimpl();
		}
		x_fprintf(out, "\t%s %s, %s, %s\n",
			p, dest[0]->name, dest[0]->name,
			src->src_pregs[0]->name);
		return;
	} else {
		int	is_64bit = 0;

		if ((IS_LLONG(ty->code) && backend->abi != ABI_POWER64)
			/*|| formod*/) {
			/* nwcc_(u)lldiv(dest, src, for_mod, data_size) */
			char	*func;

			if (ty->sign) {
				func = "__nwcc_lldiv";
			} else {
				func = "__nwcc_ulldiv";
			}
			do_softarit_call(src, func, TOK_OP_DIVIDE, formod);
		} else {
			char	*instr = NULL;

			if (ty->sign == TOK_KEY_UNSIGNED) {
				if (backend->abi != ABI_POWER64) {
					instr = "divwu";
				} else {
					instr = "divdu";
					is_64bit = 1;
				}
			} else {
				if (backend->abi != ABI_POWER64) {
					instr = "divw";
				} else {
					instr = "divd";
					is_64bit = 1;
				}
			}
						
			if (formod) {
				/* Save destination */
				x_fprintf(out, "\tmr %s, %s\n",
					tmpgpr->name, dest[0]->name);
			}
			x_fprintf(out, "\t%s %s, %s, %s\n",
				instr,
				dest[0]->name, dest[0]->name, 
				src->src_pregs[0]->name);
			if (formod) {
				/* x - ((x / y) * y) is the remainder */
				x_fprintf(out, "\tmull%c %s, %s, %s\n",
					is_64bit? 'd': 'w',
					dest[0]->name,
					dest[0]->name,
					src->src_pregs[0]->name);
				x_fprintf(out, "\tsubfc %s, %s, %s\n",
					dest[0]->name, dest[0]->name, tmpgpr->name);
			}
		}
	}
}


static void
emit_mod(struct reg **dest, struct icode_instr *src) {
	emit_div(dest, src, 1);
}


static void
emit_mul(struct reg **dest, struct icode_instr *src) {
	struct type	*ty = src->src_vreg->type;
	char		*p = NULL;

	(void) dest;

	if (IS_FLOATING(ty->code)) {
		if (ty->code == TY_FLOAT) {
			p = "fmuls"; /* "fmul";*/
		} else if (ty->code == TY_DOUBLE) {
			p = "fmul";
		} else {
			unimpl();
		}
		x_fprintf(out, "\t%s %s, %s, %s\n",
			p, dest[0]->name, dest[0]->name,
			src->src_pregs[0]->name);
		return;
	} else {
		char	*func;

		if (IS_LLONG(ty->code) && backend->abi != ABI_POWER64) {
			if (src->src_vreg->type->sign == TOK_KEY_UNSIGNED) {
				func = "__nwcc_ullmul";
			} else {
				func = "__nwcc_llmul";
			}
			do_softarit_call(src, func, TOK_OP_MULTI, 0);
		} else {
			char *instr = NULL;

			if ((IS_LONG(ty->code) && backend->abi != ABI_POWER64)
				|| IS_INT(ty->code)) {
				instr = "mullw";
			} else {
				instr = "mulld";
			}
			x_fprintf(out, "\t%s %s, %s, %s\n",
				instr, dest[0]->name, dest[0]->name,
				src->src_pregs[0]->name);
		}
	}
}

/* XXX slawi */
static void
emit_shl(struct reg **dest, struct icode_instr *src) {
	char	*p;
	(void) dest; (void) src;
	
	p = src->dest_vreg->size == 8? "sld": "slw";
	if (IS_LLONG(src->dest_vreg->type->code)
		&& backend->abi != ABI_POWER64) {
		do_softarit_call(src, "__nwcc_shift_left", TOK_OP_BSHL, 0);
		return;
	}
	if (src->src_vreg->from_const) {
		emit_load(tmpgpr, src->src_vreg);
	}
	x_fprintf(out, "\t%s %s, %s, ", p, dest[0]->name, dest[0]->name);
	if (src->src_vreg->from_const) {
		x_fprintf(out, "%s\n", tmpgpr->name);
	} else {
		x_fprintf(out, "%s\n", src->src_pregs[0]->name);
	}
}

/* XXX srawi ? */
static void
emit_shr(struct reg **dest, struct icode_instr *src) {
	char	*p;

	(void) dest; (void) src;
	p = src->dest_vreg->size == 8? "srd": "srw";
	if (IS_LLONG(src->dest_vreg->type->code)
		&& backend->abi != ABI_POWER64) {
		do_softarit_call(src, "__nwcc_shift_right", TOK_OP_BSHR, 0);
		return;
	}
	if (src->src_vreg->from_const) {
		emit_load(tmpgpr, src->src_vreg);
	}
	x_fprintf(out, "\t%s %s, %s, ", 
		p, dest[0]->name, dest[0]->name);
	if (src->src_vreg->from_const) {
		x_fprintf(out, "%s\n", tmpgpr->name);
	} else {	
		x_fprintf(out, "%s\n", src->src_pregs[0]->name);
	}
}

static void
emit_or(struct reg **dest, struct icode_instr *src) {
	char	*p;

	if (src->dest_vreg->size == 8) {
		if (IS_LLONG(src->src_vreg->type->code)
			&& backend->abi != ABI_POWER64) {
			x_fprintf(out, "\tor %s, %s, %s\n",
				dest[0]->name, dest[0]->name,
				src->src_pregs[0]->name);
			x_fprintf(out, "\tor %s, %s, %s\n",
				dest[1]->name, dest[1]->name,
				src->src_pregs[1]->name);
			return;
		} else {
			p = "or";
		}
	} else {
		p = "or";
	}	
	x_fprintf(out, "\t%s %s, %s, %s\n",
		p, dest[0]->name, dest[0]->name, src->src_pregs[0]->name);
}

static void
emit_preg_or(struct reg **dest, struct icode_instr *src) {
	x_fprintf(out, "\tor %s, %s, %s\n", dest[0]->name, dest[0]->name,
		src->src_pregs[0]->name);
}


static void
emit_and(struct reg **dest, struct icode_instr *src) {
	char	*p;

	if (src->dest_vreg->size == 8) {
		if (IS_LLONG(src->dest_vreg->type->code)
			&& backend->abi != ABI_POWER64) {
			x_fprintf(out, "\tand %s, %s, %s\n",
				dest[0]->name, dest[0]->name,
				src->src_vreg?
					 src->src_pregs[0]->name:
					"0");
			x_fprintf(out, "\tand %s, %s, %s\n",
				dest[1]->name, dest[1]->name,
				src->src_vreg?
					src->src_pregs[1]->name:
					"0");
			return;
		} else {
			p = "and";
		}
	} else {
		p = "and";
	}	
	x_fprintf(out, "\t%s %s, %s, %s\n",
		p, dest[0]->name, dest[0]->name,
		src->src_vreg? src->src_pregs[0]->name: "0");
}

static void
emit_xor(struct reg **dest, struct icode_instr *src) {
	if (IS_LLONG(src->src_vreg->type->code)
		&& backend->abi != ABI_POWER64) {
		x_fprintf(out, "\txor %s, %s, %s\n",
			dest[0]->name, dest[0]->name,
			src->src_pregs[0]->name);
		x_fprintf(out, "\txor %s, %s, %s\n",
			dest[1]->name, dest[1]->name,
			src->src_pregs[1]->name);
		return;
	}
	x_fprintf(out, "\txor %s, %s, ", 
		dest[0]->name, dest[0]->name);
	if (src->src_vreg == NULL) {
		x_fprintf(out, "%s\n", dest[0]->name);
	} else {	
		print_mem_or_reg(src->src_pregs[0], src->src_vreg);
		x_fputc('\n', out);
	}
}

/*
 * XXX long long support!!!! 
 */
static void
emit_not(struct reg **dest, struct icode_instr *src) {
	char	*p;

	/*
	 * NEG means flip all bits, then add one. NOT means flip
	 * all bits. As there does not seem to be a NOT instruction,
	 * let's use NEG, then SUB 1
	 */
	(void) src;
	x_fprintf(out, "\tneg %s, %s\n", dest[0]->name, dest[0]->name);
	p = "subic";
	x_fprintf(out, "\t%s %s, %s, 1\n",
		p, dest[0]->name, dest[0]->name);	
}	

static void
emit_ret(struct icode_instr *ii) {
	(void) ii;

	if (backend->abi == ABI_POWER64) {
		x_fprintf(out, "\tld 0, 16(1)\n"); /*XXX correct? */
	} else {
		x_fprintf(out, "\tlwz 0, 8(1)\n");
	}	
	x_fprintf(out, "\tmtlr 0\n");
/*	x_fprintf(out, "\tlwz 31, -4(1)\n");*/
	x_fprintf(out, "\tblr\n");
}

static struct icode_instr	*last_cmp_instr;


static void
emit_cmp(struct reg **dest, struct icode_instr *src) {
	last_cmp_instr = src;

	(void) src; (void) dest;
}

static void
emit_branch(struct icode_instr *ii) {
	char		*lname;
	struct vreg	*src_vreg = NULL;
	struct reg	**src_pregs = NULL;
	struct reg	**dest_pregs = NULL;
	struct reg	*src_preg;
	struct type	*cmpty = NULL;
	int		is_signed = 0;
	char		kname[256];
	static int	was_llong;
	static unsigned long	kludge;

	if (last_cmp_instr) {
		/* Not unconditional jump - get comparison */
		src_vreg = last_cmp_instr->src_vreg;
		src_pregs = last_cmp_instr->src_pregs;
		dest_pregs = last_cmp_instr->dest_pregs;
		cmpty = last_cmp_instr->dest_vreg->type;

#if 0 
		is_signed =
			last_cmp_instr->dest_vreg->type->sign
				!= TOK_KEY_UNSIGNED;
#endif
		is_signed = cmpty->sign != TOK_KEY_UNSIGNED;
	}

	if (src_pregs == NULL || src_vreg == NULL) {
		/* Compare with zero. XXX not for fp!!! */
		x_fprintf(out, "\tli %s, 0\n", power_gprs[0].name);
		src_preg = &power_gprs[0];
	} else {
		src_preg = src_pregs[0];
	}

	lname = ((struct icode_instr *)ii->dat)->dat;

	if (ii->type != INSTR_JUMP) {
		char	*opcode;

		if (src_preg->type == REG_FPR) {
			opcode = "fcmpu";
		} else {
			if (backend->abi == ABI_POWER64
				&& cmpty != NULL
				&& (IS_LONG(cmpty->code)
				|| IS_LLONG(cmpty->code))) {
				if (is_signed) {
					opcode = "cmpd";
				} else {
					opcode = "cmpld";
				}
			} else {
				if (is_signed) {
					opcode = "cmpw";
				} else {
					opcode = "cmplw";
				}
			}
		}	
#if 0
		x_fprintf(out, "\t%s 7, %s, %s\n",
			src_preg->type == REG_FPR? "fcmpu": "cmpw",
			dest_pregs[0]->name, src_preg->name);
#endif
		x_fprintf(out, "\t%s 7, %s, %s\n",
			opcode, dest_pregs[was_llong]->name,
			src_preg == &power_gprs[0]? src_preg->name:
			src_pregs[was_llong]->name);	
		if (was_llong) {
			was_llong = 0;
		} else if (last_cmp_instr->dest_vreg->is_multi_reg_obj) {
			was_llong = 1;
		}
	}

	/*
	 * XXX Most branch instructions can only deal with 16bit
	 * (or something) offsets. So in order to avoid these
	 * problems, we use ``b'' instead, which can address
	 * 24 or so. We should count the instructions between
	 * branch and target label, at least heuristically, to
	 * use normal branches for short distances in the
	 * future
	 */
	sprintf(kname, ".kludge%lu", kludge++);
	switch (ii->type) {
	case INSTR_JUMP:
		x_fprintf(out, "\tb .%s\n", lname);
		break;
	case INSTR_BR_EQUAL:
#if 0
		x_fprintf(out, "\tbeq 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tbne 7, %s\n", kname);
		break;
	case INSTR_BR_SMALLER:
#if 0
		x_fprintf(out, "\tblt 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tbge 7, %s\n", kname);
		break;
	case INSTR_BR_SMALLEREQ:
#if 0
		x_fprintf(out, "\tble 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tbgt 7, %s\n", kname);
		break;
	case INSTR_BR_GREATER:
#if 0
		x_fprintf(out, "\tbgt 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tble 7, %s\n", kname);
		break;
	case INSTR_BR_GREATEREQ:
#if 0
		x_fprintf(out, "\tbge 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tblt 7, %s\n", kname);
		break;
	case INSTR_BR_NEQUAL:
#if 0
		x_fprintf(out, "\tbne 7, %s\n",
			make_label_name(lname));
#endif
		x_fprintf(out, "\tbeq 7, %s\n", kname);
		break;
	}
	if (ii->type != INSTR_JUMP) {
		x_fprintf(out, "\tnop\n");
		x_fprintf(out, "\tb .%s\n", lname);
		x_fprintf(out, "%s:\n", kname);
	}
	last_cmp_instr = NULL; /* XXX hmm ok? */
	x_fprintf(out, "\tnop\n"); /* fill branch delay slot */
}


static void
emit_mov(struct copyreg *cr) {
	struct reg	*dest = cr->dest_preg;
	struct reg	*src = cr->src_preg;

	if (src == NULL) {
		/* Move null to register */
		x_fprintf(out, "\tmr %s, $0\n", /* XXX !? */
			dest->name);
	} else {
		if (dest->type != src->type) {
			/* Must be fpr vs gpr */
			unimpl();
#if 0
			/* XXX FP use fctiw!! */
#endif
		} else if (dest->type == REG_FPR) {
			x_fprintf(out, "\tfmr %s, %s\n",
				dest->name, src->name);
		} else {
			x_fprintf(out, "\tmr %s, %s\n",
				dest->name, src->name);
		}
	}	
}


static void
emit_setreg(struct reg *dest, int *value) {
	/* XXX cross-comp */
	do_emit_li(dest->name, *(unsigned *)value, 0); /* XXX 1 */
}

static void
emit_xchg(struct reg *r1, struct reg *r2) {
	(void) r1; (void) r2;
	unimpl();
}


static void
emit_addrof(struct reg *dest, struct vreg *src, struct vreg *structtop) {
	struct decl	*d;
	long		offset = 0;
	char		*sign = NULL;

	if (src == NULL) {
		d = curfunc->proto->dtype->tlist->tfunc->lastarg;
	} else {
		d = src->var_backed;
	}	

	if (structtop != NULL) {
		d = structtop->var_backed;
	}	

	if (d && d->stack_addr != NULL) {
		if (d->stack_addr->is_func_arg) {
			sign = "+";
		} else {
			sign = "-";
		}	
		offset = d->stack_addr->offset;
	}	

	if (src && src->parent != NULL) {
		/* Structure or union type */
		if (d != NULL) {
			if (d->stack_addr != NULL) {
				x_fprintf(out, "\taddi %s, %s, %lu\n",
					dest->name,
					d->stack_addr->use_frame_pointer?
						"31": "1",
					offset+calc_offsets(src));
			} else {
				/* Static */
				x_fprintf(out, "\t%s %s, _Toc_%s(2)\n",
					backend->abi == ABI_POWER64? "ld": "lwz",	
					dest->name, d->dtype->name);
#if 0
				x_fprintf(out, "\taddi %s, %s, %lu\n",
					dest->name, dest->name,
					calc_offsets(src)); 
#endif
				do_addi(dest, dest, calc_offsets(src));
			}	
		} else if (structtop->from_ptr) {
#if 0
			x_fprintf(out, "\taddi %s, %s, %lu\n",
				dest->name,
				structtop->from_ptr->pregs[0]->name,
				calc_offsets(src));
#endif
			do_addi(dest, structtop->from_ptr->pregs[0],
				calc_offsets(src));
		} else {
			printf("hm attempt to take address of %s\n",
				src->type->name);	
			unimpl();
		}	
	} else if (src && src->from_ptr) {	
		x_fprintf(out, "\tmr %s, %s\n",
			dest->name, src->from_ptr->pregs[0]->name);
	} else {		
		if (d && d->stack_addr) {
			/* src = NULL means start of variadic */
			x_fprintf(out, "\taddi %s, %s, %ld\n",
				dest->name,
				d->stack_addr->use_frame_pointer? "31": "1",
				offset);
		} else if (d) {
			/*
			 * Must be static variable - symbol itself is
			 * address
			 */
			x_fprintf(out, "\t%s %s, %s%s(2)\n",
				backend->abi == ABI_POWER64? "ld": "lwz",	
				dest->name,
				d->dtype->tlist && d->dtype->tlist->
				type == TN_FUNCTION? "_Toc_": "_Toc_",
				d->dtype->name);
		} else if (src && src->stack_addr) {
			/* Anonymous item */
			x_fprintf(out, "\taddi %s, %s, %lu\n",
				dest->name,
				src->stack_addr->use_frame_pointer?
				"31": "1",
				(unsigned long)src->stack_addr->offset);
		} else {
			unimpl(); /* ??? */
		}
	}	
}


/*
 * Copy initializer to automatic variable of aggregate type
 */
static void
emit_copyinit(struct decl *d) {
	x_fprintf(out, "\taddi 3, 31, %lu\n", d->stack_addr->offset);
	x_fprintf(out, "\t%s 4, %s(2)\n",
		backend->abi == ABI_POWER64? "ld": "lwz",
		d->init_name->name);
	/* XXX dli?? */
	do_emit_li("5", d->vreg->size, 0);
	x_fprintf(out, "\tbl .memcpy\n");
	x_fprintf(out, "\tnop\n");
}


/*
 * Assign one struct to another (may be any of automatic or static or
 * addressed thru pointer)
 */
static void
emit_copystruct(struct copystruct *cs) {
	struct vreg	*stop;

	/* XXX dli?! */
	do_emit_li("5", (unsigned)cs->src_vreg->size, 0);
	if (cs->src_from_ptr == NULL) {
		if (cs->src_vreg->parent) {
			stop = get_parent_struct(cs->src_vreg);
		} else {
			stop = NULL;
		}	
		emit_addrof(tmpgpr, cs->src_vreg, stop); 
		x_fprintf(out, "\tmr 4, %s\n", tmpgpr->name);
	} else {
		if (cs->src_vreg->parent) {
			x_fprintf(out, "\t%siu %s, %s, %lu\n",
				backend->abi == ABI_POWER64? /*"dadd"*/"add": "add",
				cs->src_from_ptr->name,
				cs->src_from_ptr->name,
				calc_offsets(cs->src_vreg));
		}
		x_fprintf(out, "\tmr 4, %s\n", cs->src_from_ptr->name);
	}

	if (cs->dest_vreg == NULL) {
		/* Copy to hidden return pointer */
		emit_load(tmpgpr, curfunc->hidden_pointer);
		x_fprintf(out, "\tmr 3, %s\n", tmpgpr->name);
	} else if (cs->dest_from_ptr == NULL) {
		if (cs->dest_vreg->parent) {
			stop = get_parent_struct(cs->dest_vreg);
		} else {
			stop = NULL;
		}	
		emit_addrof(tmpgpr, cs->dest_vreg, stop); 
		x_fprintf(out, "\tmr 3, %s\n", tmpgpr->name);
	} else {	
		if (cs->dest_vreg->parent) {
			x_fprintf(out, "\t%siu %s, %s, %lu\n",
				backend->abi == ABI_POWER64? /*"dadd"*/"add": "add",
				cs->dest_from_ptr->name,
				cs->dest_from_ptr->name,
				calc_offsets(cs->dest_vreg));
		}	
		x_fprintf(out, "\tmr 3, %s\n", cs->dest_from_ptr->name);
	}
	x_fprintf(out, "\tbl .memcpy\n");
	x_fprintf(out, "\tnop\n");
}

static void
emit_zerostack(struct stack_block *sb, size_t nbytes) {
	static struct vreg	vr;

	x_fprintf(out, "\tli 5, %lu\n", /* XXX dli!??! */
		(unsigned long)nbytes);	
	x_fprintf(out, "\tli 4, 0\n");
	vr.stack_addr = sb;
	emit_addrof(&power_gprs[3], &vr, NULL);
	x_fprintf(out, "\tbl .memset\n");
	x_fprintf(out, "\tnop\n");
}

static void
emit_alloca(struct allocadata *ad) {
	x_fprintf(out, "\tmr 3, %s\n", ad->size_reg->name);
	x_fprintf(out, "\tbl .malloc\n");
	x_fprintf(out, "\tnop\n");
	if (&power_gprs[3] != ad->result_reg) {
		x_fprintf(out, "\tmove %s, 3\n",
			ad->result_reg->name);
	}	
}	

static void
emit_dealloca(struct stack_block *sb) {
	x_fprintf(out, "\t%s 3, %lu(31)\n",
		backend->abi == ABI_POWER64? "ld": "lwz", sb->offset);
	x_fprintf(out, "\tbl .free\n");
	x_fprintf(out, "\tnop\n");
}	

static void
do_stack(FILE *out, struct decl *d, unsigned long offset) {
	if (d->stack_addr->use_frame_pointer) {
		x_fprintf(out, "%lu(31)", d->stack_addr->offset+offset);
	} else {
		x_fprintf(out, "%lu(1)", d->stack_addr->offset+offset);
	}
}

static void
print_mem_operand(struct vreg *vr, struct token *constant,
struct vreg *vr_parent) { 
	static int	was_llong;

	if (vr && vr->from_const != NULL) {
		constant = vr->from_const;
	}
	if (constant != NULL) {
		struct token	*t = vr->from_const;

		/*if (t->type == TY_INT) {
			x_fprintf(out, "%d", 
				*(int *)t->data);
		} else if (t->type == TY_UINT) {
			x_fprintf(out, "%u", 
				*(unsigned *)t->data);
		}*/
		
		if (IS_INT(t->type)) {
			cross_print_value_by_type(out, t->data, t->type, 0);
		} else if (backend->abi == ABI_POWER64
			&& (IS_LLONG(t->type)	
			|| IS_LONG(t->type))) {	
			struct ty_llong	*tll = t->data2;

			if (tll == NULL) {
				/* Small value - can be immediate */
				unimpl();
			} else {	
				x_fprintf(out, "_Llong%lu(2)",
					tll->count);	
			}	
			return;
		} else if (IS_LONG(t->type)) {
			cross_print_value_by_type(out, t->data, t->type, 0);
#if 0
		} else if (t->type == TY_LONG) {
			x_fprintf(out, "%ld", 
				*(long *)t->data);
		} else if (t->type == TY_ULONG) {
			x_fprintf(out, "%lu",
				*(unsigned long *)t->data);
#endif
		} else if (IS_LLONG(t->type)) {
			char		*p;
			static int	was_llong;

			if (was_llong) {
				/* Loading second part of long long */
				p = (char *)t->data + 4;
				was_llong = 0;
			} else {
				p = t->data;
				was_llong = 1;
			}
#if 0
			x_fprintf(out, "0x%x", *(unsigned *)p);
#endif
			cross_print_value_by_type(out, t->data, TY_UINT, 0);
		} else if (t->type == TOK_STRING_LITERAL) {
			struct ty_string	*ts = t->data;
			x_fprintf(out, "_Str%ld(2)",
				ts->count);
		} else if (t->type == TY_FLOAT
			|| t->type == TY_DOUBLE
			|| t->type == TY_LDOUBLE) {
			struct ty_float	*tf = t->data;
	
			x_fprintf(out, "_Float%lu(2)",
				tf->count);
		} else {
			printf("loadimm: Bad data type %d\n", t->type);
			exit(EXIT_FAILURE);
		}
	} else if (vr->parent != NULL) {
		struct vreg	*vr2 = vr_parent;
		struct decl	*d2;

		if ((d2 = vr2->var_backed) != NULL) {
			if (d2->stack_addr) {
				do_stack(out,
					vr2->var_backed,
					calc_offsets(vr)
					+ (was_llong? 4: 0));
			} else {
				/* static */
				x_fprintf(out, "%lu",
#if 0
					calc_offsets(vr)
#endif
					/*
					 * XXX static structs require a
					 * TOC entry in PowerOpen so the
					 * preferred way of loading a
					 * member now is
					 * lwz 13, static_struct(2)
					 * addi 13, 13, offset
					 * lwz dest, 0(13)
					 * rather than
				 	 * ... lwz dest, offset(13)
					 */
					0
					+ (was_llong? 4: 0));
				x_fprintf(out, "(%s)",
					tmpgpr->name);
			}
		} else if (vr2->from_ptr) {
			/* Struct comes from pointer */
			x_fprintf(out, "%lu",
				(unsigned long)calc_offsets(vr)
				+ (was_llong? 4: 0));
			x_fprintf(out, "(%s)",
				vr2->from_ptr->pregs[0]->name);	
		}else {
			printf("BUG: Bad load for %s\n",
				vr->type->name? vr->type->name: "structure");
			abort();
		}	
	} else if (vr->var_backed) {
		struct decl	*d = vr->var_backed;

		if (d->stack_addr != NULL) {
			do_stack(out, d, (was_llong? 4: 0));
		} else {
			/*
			 * Static or register variable
			 */
			if (d->dtype->storage == TOK_KEY_REGISTER) {
				unimpl();
			} else {
#if 0
				x_fprintf(out, "0(%s)", tmpgpr->name);
#endif
				x_fprintf(out, "%s%s(2)",
					vr->var_backed->dtype->tlist
			&& vr->var_backed->dtype->tlist->type == TN_FUNCTION? 
					"_Toc_": "_Toc_",
					vr->var_backed->dtype->name);
			}	
		}
	} else if (vr->stack_addr) {
		unsigned long	offset = vr->stack_addr->offset +
			(was_llong? 4: 0);
		if (vr->stack_addr->use_frame_pointer) {
			x_fprintf(out, "%lu(31)", offset);
		} else {
			x_fprintf(out, "%lu(1)", offset);
		}
	} else if (vr->from_ptr) {
		x_fprintf(out, "%d(%s)",
			(was_llong? 4: 0),
			vr->from_ptr->pregs[0]->name);
	} else {
		abort();
	}
	if (constant == NULL) {
		if (was_llong) {
			/*x_fprintf(out, " + 4 ");*/
			was_llong = 0;
		} else if (vr->is_multi_reg_obj) {
			was_llong = 1;
		}	
	}	
}

static void
emit_power_srawi(struct icode_instr *ii) {
	x_fprintf(out, "\tsrawi %s, %s, %d\n",
		ii->dest_pregs[0]->name,
		ii->src_pregs[0]->name,
		*(int *)ii->dat);
}

static void
emit_power_rldicl(struct icode_instr *ii) {
	x_fprintf(out, "\trldicl %s, %s, %s, %d\n",
		ii->dest_pregs[0]->name,
		ii->dest_pregs[0]->name,
		ii->src_pregs[0]->name,
		*(int *)ii->dat);
}

static void
emit_power_fcfid(struct icode_instr *ii) {
	x_fprintf(out, "\tfcfid %s, %s\n",
		ii->dest_pregs[0]->name,
		ii->src_pregs[0]->name);
}

static void
emit_power_frsp(struct icode_instr *ii) {
	struct reg	*r = ii->dat;
	x_fprintf(out, "\tfrsp %s, %s\n",
		r->name, r->name);
}

static void
emit_power_rlwinm(struct icode_instr *ii) {
	x_fprintf(out, "\trlwinm %s, %s, %s, 0x%x\n",
		ii->dest_pregs[0]->name,
		ii->dest_pregs[0]->name,
		ii->src_pregs[0]->name,
		*(int *)ii->dat);
}

static void
emit_power_slwi(struct icode_instr *ii) {
	x_fprintf(out, "\tslwi %s, %s, %d\n",
		ii->dest_pregs[0]->name,
		ii->src_pregs[0]->name,
		*(int *)ii->dat);
}

static void
emit_power_extsb(struct icode_instr *ii) {
	struct reg	*r = ii->dat;

	x_fprintf(out, "\textsb %s, %s\n", r->name, r->name);
}

static void
emit_power_extsh(struct icode_instr *ii) {
	struct reg	*r = ii->dat;

	x_fprintf(out, "\textsh %s, %s\n", r->name, r->name);
}

static void
emit_power_extsw(struct icode_instr *ii) {
	struct reg	*r = ii->dat;

	x_fprintf(out, "\textsw %s, %s\n", r->name, r->name);
}

static void
emit_power_xoris(struct icode_instr *ii) {
	struct reg	*dest = ii->dest_pregs[0];

	x_fprintf(out, "\txoris %s, %s, 0x%x\n",
		dest->name, dest->name, *(int *)ii->dat);
}

static void
emit_power_lis(struct icode_instr *ii) {
	x_fprintf(out, "\tlis %s, 0x%x\n",
		ii->dest_pregs[0]->name, *(int *)ii->dat);
}

static void
emit_power_loadup4(struct icode_instr *ii) {
	x_fprintf(out, "\tlwz %s, %lu(31)\n",
		ii->dest_pregs[0]->name,
		(unsigned long)ii->src_vreg->stack_addr->offset+4);
}

static void
emit_power_fctiwz(struct icode_instr *ii) {
	struct reg	*r = ii->dat;

	x_fprintf(out, "\tfrsp %s, %s\n", r->name, r->name);
	x_fprintf(out, "\tfctiwz %s, %s\n", r->name, r->name);
}

struct emitter power_emit_as = {
	init,
	emit_strings,
	emit_extern_decls,
	emit_static_decls,
	emit_struct_defs,
	emit_comment,
	NULL, /* emit_dwarf2_line */
	NULL, /* emit_dwarf2_files */
	emit_inlineasm,
	emit_unimpl,
	emit_empty,
	emit_label,
	emit_call,
	emit_callindir,
	emit_func_intro,
	emit_func_outro,
	NULL,
	NULL,
	emit_allocstack,
	emit_freestack,
	emit_adj_allocated,
	emit_inc,
	emit_dec,
	emit_load,
	emit_store,
	emit_setsect,
	emit_alloc,
	emit_neg,
	emit_sub,
	emit_add,
	emit_div,
	emit_mod,
	emit_mul,
	emit_shl,
	emit_shr,
	emit_or,
	emit_preg_or,
	emit_and,
	emit_xor,
	emit_not,
	emit_ret,
	emit_cmp,
	NULL, /* extend_sign */
	NULL, /* conv_fp */
	emit_branch,
	emit_mov,
	emit_setreg,
	emit_xchg,
	emit_addrof,
	emit_copyinit,
	emit_copystruct,
	emit_zerostack,
	emit_alloca,
	emit_dealloca,
	emit_struct_inits,
	NULL,
	NULL,
	NULL /* print_mem_operand */
};

struct emitter_power power_emit_power_as = {
	emit_power_srawi,
	emit_power_rldicl,
	emit_power_fcfid,
	emit_power_frsp,
	emit_power_rlwinm,
	emit_power_slwi,
	emit_power_extsb,
	emit_power_extsh,
	emit_power_extsw,
	emit_power_xoris,
	emit_power_lis,
	emit_power_loadup4,
	emit_power_fctiwz
};	

