#ifndef TYPEMAP_H
#define TYPEMAP_H

#include <stdio.h> /* size_t */

struct token;
struct tyval;
struct expr;
struct type;


/*
 * This does not take stuff like PDP-endianness into account,
 * and also assumes that all data pointers are the same size
 */
struct arch_properties {
	int	arch;
	int	abi;
	int	endianness;
#define ENDIAN_BIG	1
#define ENDIAN_LITTLE	2
	int	data_ptr_size;
	int	func_ptr_size;
};

/*
 * If the number of bits in a host and a target type equal,
 * and the sign too, then we assume the types are compatible
 * for all arithmetic operations
 */
struct type_properties {
	int	code;
	int	sign;		/* sign */ 
	int	bytes;		/* size in bytes */
	int	align;		/* natural alignment in bytes */
	int	bits;		/* size in bits */
	char	*max_dec;	/* max range in decimal ascii */
	int	max_dec_len;	/* length of max_dec */
	char	*max_hex;	/* ... etc ..  */
	int	max_hex_len;
	char	*max_oct;
	int	max_oct_len;
};


/*
 * Properties of a target architecture. Not all of it is used yet,
 * but should be sooner or later.
 *
 * type_info points to an array containing type information. 
 * arch_info is a grabbag of other architecture stuff, including
 * some type info too (general pointer size)
 */
struct target_properties {
	struct type_properties	*type_info;
	struct arch_properties	*arch_info;
};

struct num;

void	cross_initialize_type_map(int targetarch, int targetabi);
struct type_properties *cross_get_type_properties(int code);
int	cross_exec_op(struct token *tok, struct tyval *res, struct tyval *dest,
		struct tyval *src);
void	cross_do_conv(struct tyval *dest, int code);
int	cross_convert_tyval(struct tyval *left, struct tyval *right);
size_t	cross_get_sizeof_type(struct type *);
struct num *cross_scan_value(const char *str, int type,
		int hexa_flag, int octal_flag, int fp_flag);
void	cross_print_value_by_type(FILE *out, void *value, int type, int ofmt);
void	cross_print_value_chunk(FILE *out, void *value, int type,
		int chunk_type, int chunk_type_rem, int chunkno);
size_t	cross_to_host_size_t(struct tyval *tv);

#endif

